"""
Docstring parser
"""

from cdd.shared.docstring_parsers import parse_docstring


def docstring(
    doc_string,
    infer_type=False,
    return_tuple=False,
    parse_original_whitespace=False,
    emit_default_prop=True,
    emit_default_doc=True,
):
    """
    Converts a docstring to an AST

    :param doc_string: docstring portion
    :type doc_string: ```Union[str, Dict]```

    :param infer_type: Whether to try inferring the typ (from the default)
    :type infer_type: ```bool```

    :param return_tuple: Whether to return a tuple, or just the intermediate_repr
    :type return_tuple: ```bool```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :param emit_default_prop: Whether to include the default dictionary property.
    :type emit_default_prop: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :return: intermediate_repr, whether it returns or not
    :rtype: ```Optional[Union[dict, Tuple[dict, bool]]]```
    """
    assert isinstance(
        doc_string, str
    ), "Expected `str` got `{doc_string_type!r}`".format(
        doc_string_type=type(doc_string).__name__
    )
    parsed = (
        doc_string
        if isinstance(doc_string, dict)
        else parse_docstring(
            doc_string,
            infer_type=infer_type,
            emit_default_prop=emit_default_prop,
            emit_default_doc=emit_default_doc,
            parse_original_whitespace=parse_original_whitespace,
        )
    )

    if return_tuple:
        return parsed, (
            "returns" in parsed
            and parsed["returns"] is not None
            and "return_type" in (parsed.get("returns") or iter(()))
        )

    return parsed


__all__ = ["docstring"]



"""Mocks for cst"""

from cdd.shared.cst_utils import (
    AnnAssignment,
    Assignment,
    AugAssignment,
    CallStatement,
    ClassDefinitionStart,
    CommentStatement,
    ElifStatement,
    ElseStatement,
    ExprStatement,
    FalseStatement,
    FromStatement,
    FunctionDefinitionStart,
    IfStatement,
    NoneStatement,
    PassStatement,
    ReturnStatement,
    TripleQuoted,
    TrueStatement,
    UnchangingLine,
)

cstify_cst = (
    CommentStatement(line_no_start=1, line_no_end=1, value="# pragma: no cover"),
    CommentStatement(line_no_start=1, line_no_end=2, value="\n# flake8: noqa"),
    CommentStatement(
        line_no_start=2,
        line_no_end=3,
        value="\n# === Copyright 2022 under CC0",
    ),
    TripleQuoted(
        is_double_q=True,
        is_docstr=False,
        line_no_start=3,
        line_no_end=4,
        value='\n"""Module docstring goes here"""',
    ),
    FromStatement(line_no_start=4, line_no_end=6, value="\n\nfrom operator import add"),
    ClassDefinitionStart(
        line_no_start=6, line_no_end=9, value="\n\n\nclass C(object):", name="C"
    ),
    TripleQuoted(
        is_double_q=True,
        is_docstr=True,
        line_no_start=9,
        line_no_end=10,
        value='\n    """My cls"""',
    ),
    FunctionDefinitionStart(
        line_no_start=10,
        line_no_end=13,
        value="\n" "\n" "    @staticmethod\n" "    def add1(foo):",
        name="add1",
    ),
    TripleQuoted(
        is_double_q=True,
        is_docstr=True,
        line_no_start=13,
        line_no_end=20,
        value='\n        """\n'
        "        :param foo: a foo\n"
        "        :type foo: ```int```\n"
        "\n"
        "        :return: foo + 1\n"
        "        :rtype: ```int```\n"
        '        """',
    ),
    TripleQuoted(
        is_double_q=True,
        is_docstr=False,
        line_no_start=20,
        line_no_end=22,
        value='\n\n        """foo"""',
    ),
    FunctionDefinitionStart(
        line_no_start=22, line_no_end=23, value="\n        def g():", name="g"
    ),
    TripleQuoted(
        is_double_q=True,
        is_docstr=True,
        line_no_start=23,
        line_no_end=23,
        value=' """foo : bar ; can"""',
    ),
    PassStatement(line_no_start=23, line_no_end=23, value="; pass"),
    FunctionDefinitionStart(
        line_no_start=23, line_no_end=25, value="\n\n        def h():", name="h"
    ),
    CommentStatement(line_no_start=25, line_no_end=25, value=" # stuff"),
    PassStatement(line_no_start=25, line_no_end=26, value="\n            pass"),
    FunctionDefinitionStart(
        line_no_start=26,
        line_no_end=29,
        value="\n\n        def adder(a: int,\n                  b: int) -> int:",
        name="adder",
    ),
    TripleQuoted(
        is_double_q=True,
        is_docstr=True,
        line_no_start=29,
        line_no_end=36,
        value="\n"
        '            """\n'
        "            :param a: First arg\n"
        "\n"
        "            :param b: Second arg\n"
        "\n"
        "            :return: first + second arg\n"
        '            """',
    ),
    CommentStatement(
        line_no_start=36, line_no_end=37, value="\n            # fmt: off"
    ),
    AnnAssignment(
        line_no_start=37,
        line_no_end=40,
        value="\n            res: \\\n                int \\\n                = a + b",
    ),
    ReturnStatement(line_no_start=40, line_no_end=41, value="\n            return res"),
    Assignment(
        line_no_start=41,
        line_no_end=47,
        value="\n\n        r = (\n            add(foo, 1)\n            or\n            adder(foo, 1)\n        )",
    ),
    IfStatement(line_no_start=47, line_no_end=48, value="\n        if r:"),
    NoneStatement(line_no_start=48, line_no_end=49, value="\n            None"),
    ElifStatement(line_no_start=49, line_no_end=50, value="\n        elif r:"),
    TrueStatement(line_no_start=50, line_no_end=51, value="\n            True"),
    FalseStatement(line_no_start=51, line_no_end=52, value="\n            False"),
    CommentStatement(
        line_no_start=52, line_no_end=53, value="\n            # ([5,5] @ [5,5]) *\\"
    ),
    ExprStatement(
        line_no_start=53,
        line_no_end=55,
        value="\n            -5 / 7 ** 6 + \\\n            6.0 - 6e1 & 1+2.34j",
    ),
    AugAssignment(line_no_start=55, line_no_end=56, value="\n            r <<= 5"),
    CallStatement(line_no_start=56, line_no_end=57, value="\n            print(r)"),
    ElseStatement(line_no_start=57, line_no_end=58, value="\n        else:"),
    PassStatement(line_no_start=58, line_no_end=59, value="\n            pass"),
    CommentStatement(line_no_start=59, line_no_end=60, value="\n        # fmt: on"),
    CommentStatement(
        line_no_start=60,
        line_no_end=61,
        value="\n        # That^ incremented `foo` by 1",
    ),
    ReturnStatement(line_no_start=61, line_no_end=62, value="\n        return r"),
    CommentStatement(
        line_no_start=62,
        line_no_end=65,
        value="\n\n\n# from contextlib import ContextDecorator",
    ),
    CommentStatement(
        line_no_start=65, line_no_end=67, value="\n\n# with ContextDecorator():"
    ),
    CommentStatement(line_no_start=67, line_no_end=68, value="\n#    pass"),
    FunctionDefinitionStart(
        line_no_start=68, line_no_end=71, value="\n\n\ndef f():", name="f"
    ),
    ReturnStatement(line_no_start=71, line_no_end=72, value="\n    return 1"),
    UnchangingLine(line_no_start=72, line_no_end=73, value="\n"),
)


__all__ = ["cstify_cst"]



"""
Mock routes
"""

import cdd.routes.emit.bottle
from cdd.shared.pure_utils import tab

route_config = {
    "app": "rest_api",
    "name": "Config",
    "route": "/api/config",
    "variant": -1,
}

create_route = cdd.routes.emit.bottle.create(**route_config)

route_config["primary_key"] = "dataset_name"

read_route = cdd.routes.emit.bottle.read(**route_config)
destroy_route = cdd.routes.emit.bottle.destroy(**route_config)

route_mock_prelude = (
    'rest_api = type("App", tuple(),\n'
    "{sep}{{ method: lambda h: lambda g=None: g \n"
    '{sep}  for method in ("get", "post", "put", "delete") }})\n'.format(sep=tab * 4)
)

route_prelude = (
    "from bottle import Bottle, request, response\n\n"
    "rest_api = Bottle(catchall=False, autojson=True)\n"
)

__all__ = [
    "create_route",
    "read_route",
    "destroy_route",
    "route_config",
    "route_prelude",
    "route_mock_prelude",
]



"""
Utility functions for cdd.gen
"""

import ast
from ast import Assign, ClassDef, FunctionDef, Import, ImportFrom, Name, Store
from itertools import chain
from json import load
from operator import itemgetter
from os import path

from cdd.shared.ast_utils import (
    infer_imports,
    maybe_type_comment,
    optimise_imports,
    set_value,
)
from cdd.shared.emit.utils.emitter_utils import get_emitter
from cdd.shared.parse import kind2instance_type
from cdd.shared.parse.utils.parser_utils import get_parser
from cdd.shared.pure_utils import (
    find_module_filepath,
    pascal_to_upper_camelcase,
    rpartial,
)
from cdd.shared.source_transformer import ast_parse, to_code


def get_input_mapping_from_path(emit_name, module_path, symbol_name):
    """
    Given (module_path, symbol_name) acquire file path, `ast.parse` out all top-level symbols matching `emit_name`

    :param emit_name: Which type to generate.
    :type emit_name: ```Literal["argparse", "class", "function", "json_schema",
                                "pydantic", "sqlalchemy", "sqlalchemy_table"]```

    :param module_path: Module path
    :type module_path: ```str```

    :param symbol_name: Symbol to import from module
    :type symbol_name: ```str```

    :return: Dictionary of (name, AST) where AST is produced by a cdd emitter matching `emit_name`
    :rtype: ```dict```
    """
    module_filepath = find_module_filepath(module_path, symbol_name)
    with open(module_filepath, "rt") as f:
        input_ast_mod = ast.parse(f.read())
    type_instance_must_be = kind2instance_type.get(emit_name, (FunctionDef, ClassDef))
    input_mapping = dict(
        map(
            lambda ir: (ir["name"], ir),
            map(
                lambda node: get_parser(node, emit_name)(node),
                filter(
                    rpartial(
                        isinstance,
                        type_instance_must_be,
                    ),
                    input_ast_mod.body,
                ),
            ),
        )
    )
    assert input_mapping, "Nothing parsed out of {!r}".format(module_filepath)
    return input_mapping


def get_emit_kwarg(decorator_list, emit_call, emit_name, name_tpl, name):
    """
    Emit keyword arguments have different requirements dependent on emitter
    Determine correct one, and always include the name.

    :param decorator_list: List of decorators
    :type decorator_list: ```Optional[Union[List[Str], List[]]]```

    :param emit_name: Which type to generate.
    :type emit_name: ```Literal["argparse", "class", "function", "json_schema",
                                "pydantic", "sqlalchemy", "sqlalchemy_table"]```

    :param emit_call: Whether to emit a `__call__` method from the `_internal` IR subdict
    :type emit_call: ```bool```

    :param name_tpl: Template for the name, e.g., `{name}Config`.
    :type name_tpl: ```str```

    :param name: Interpolates into `name_tpl`
    :type name: ```str```

    :return: Dictionary of keyword arguments targeted the specialised emit function.
    :rtype: ```dict``
    """
    return (
        lambda _name: {
            "argparse_function": {"function_name": _name},
            "class_": {
                "class_name": _name,
                "decorator_list": decorator_list,
                "emit_call": emit_call,
            },
            "function": {
                "function_name": _name,
            },
            "json_schema": {
                "identifier": _name,
            },
            "sqlalchemy": {"table_name": _name},
            "sqlalchemy_table": {"table_name": _name},
        }[emit_name]
    )(None if name == "infer" else name_tpl.format(name=name))


def gen_file(
    name_tpl,
    input_mapping_it,
    parse_name,
    emit_name,
    output_filename,
    prepend,
    emit_call,
    emit_and_infer_imports,
    emit_default_doc,
    decorator_list,
    no_word_wrap,
    imports,
    functions_and_classes=None,
):
    """
    Generate Python file of containing `input_mapping_it`.values converted to `emit_name`

    :param name_tpl: Template for the name, e.g., `{name}Config`.
    :type name_tpl: ```str```

    :param input_mapping_it: Import location of mapping/2-tuple collection.
    :type input_mapping_it: ```Iterator[Tuple[str,AST]]```

    :param parse_name: Which type to parse.
    :type parse_name: ```Literal["argparse", "class", "function", "json_schema",
                                 "pydantic", "sqlalchemy", "sqlalchemy_table"]```

    :param emit_name: Which type to generate.
    :type emit_name: ```Literal["argparse", "class", "function", "json_schema",
                                "pydantic", "sqlalchemy", "sqlalchemy_table"]```

    :param output_filename: Output file to write to
    :type output_filename: ```str```

    :param prepend: Prepend file with this. Use '\n' for newlines.
    :type prepend: ```Optional[str]```

    :param emit_call: Whether to emit a `__call__` method from the `_internal` IR subdict
    :type emit_call: ```bool```

    :param emit_and_infer_imports: Whether to emit and infer imports at the top of the generated code
    :type emit_and_infer_imports: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :param decorator_list: List of decorators
    :type decorator_list: ```Optional[Union[List[Str], List[]]]```

    :param no_word_wrap: Whether word-wrap is disabled (on emission).
    :type no_word_wrap: ```Optional[Literal[True]]```

    :param imports: Import to preclude in Python file
    :type imports: ```str```

    :param functions_and_classes: Functions and classes that have been preparsed
    :type functions_and_classes: ```Optional[Tuple[AST]]```
    """
    parsed_ast = gen_module(
        decorator_list,
        emit_and_infer_imports,
        emit_call,
        emit_default_doc,
        emit_name,
        functions_and_classes,
        imports,
        input_mapping_it,
        name_tpl,
        no_word_wrap,
        parse_name,
        prepend,
    )
    assert (
        len(parsed_ast.body) > 1
        or not isinstance(parsed_ast.body[0], Assign)
        and any(
            filter(
                lambda target: isinstance(target, Name) and target.id == "__all__",
                parsed_ast.body[0].targets,
            )
        )
    ), "Nothing will be append to {!r}".format(output_filename)
    with open(output_filename, "a") as f:
        f.write(to_code(parsed_ast))


def gen_module(
    decorator_list,
    emit_and_infer_imports,
    emit_call,
    emit_default_doc,
    emit_name,
    functions_and_classes,
    imports,
    input_mapping_it,
    name_tpl,
    no_word_wrap,
    parse_name,
    prepend,
    global__all__=None,
):
    """
    Generate Python module `input_mapping_it`.values converted to `emit_name`

    :param name_tpl: Template for the name, e.g., `{name}Config`.
    :type name_tpl: ```str```

    :param input_mapping_it: Import location of mapping/2-tuple collection.
    :type input_mapping_it: ```Iterator[Tuple[str,AST]]```

    :param parse_name: Which type to parse.
    :type parse_name: ```Literal["argparse", "class", "function", "json_schema",
                                 "pydantic", "sqlalchemy", "sqlalchemy_table"]```

    :param emit_name: Which type to generate.
    :type emit_name: ```Literal["argparse", "class", "function", "json_schema",
                                "pydantic", "sqlalchemy", "sqlalchemy_table"]```

    :param prepend: Prepend file with this. Use '\n' for newlines.
    :type prepend: ```Optional[str]```

    :param emit_call: Whether to emit a `__call__` method from the `_internal` IR subdict
    :type emit_call: ```bool```

    :param emit_and_infer_imports: Whether to emit and infer imports at the top of the generated code
    :type emit_and_infer_imports: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :param decorator_list: List of decorators
    :type decorator_list: ```Optional[Union[List[Str], List[]]]```

    :param no_word_wrap: Whether word-wrap is disabled (on emission).
    :type no_word_wrap: ```Optional[Literal[True]]```

    :param imports: Import to preclude in Python file
    :type imports: ```str```

    :param functions_and_classes: Functions and classes that have been preparsed
    :type functions_and_classes: ```Optional[Tuple[AST]]```

    :param global__all__: `__all__` symbols for that magic
    :type global__all__: ```List[str]```

    :return: Module with everything contained inside, e.g., all the imports, parsed out functions and classes
    :rtype: ```Module```
    """
    if global__all__ is None:
        global__all__ = []
    if functions_and_classes is None:
        functions_and_classes = get_functions_and_classes(
            decorator_list,
            emit_call,
            emit_default_doc,
            emit_name,
            global__all__,
            input_mapping_it,
            name_tpl,
            no_word_wrap,
            parse_name,
        )
    if emit_and_infer_imports:
        imports = "{}{}".format(
            imports or "",
            " ".join(
                map(
                    to_code,
                    optimise_imports(chain(*map(infer_imports, functions_and_classes))),
                )
            ),
        )

    # Too many params! - Clean things up for debugging:
    del (
        decorator_list,
        emit_call,
        emit_default_doc,
        emit_name,
        input_mapping_it,
        name_tpl,
        no_word_wrap,
        parse_name,
    )

    content = "{prepend}{imports}\n{functions_and_classes}\n{__all__}".format(
        prepend="" if prepend is None else prepend,
        imports=imports,  # TODO: Optimize imports programmatically (akin to `autoflake --remove-all-unused-imports`)
        functions_and_classes="\n\n".join(map(to_code, functions_and_classes)),
        __all__=to_code(
            Assign(
                targets=[Name("__all__", Store())],
                value=ast.parse(  # `TypeError: Type List cannot be instantiated; use list() instead`
                    str(
                        list(
                            map(
                                lambda s: s.rstrip("\n").strip("'").strip('"'),
                                map(to_code, map(set_value, global__all__)),
                            )
                        )
                    )
                )
                .body[0]
                .value,
                expr=None,
                lineno=None,
                **maybe_type_comment,
            )
        ),
    )
    parsed_ast = ast.parse(content)
    # TODO: Shebang line first, then docstring, then imports
    doc_str = ast.get_docstring(parsed_ast, clean=True)
    whole = tuple(
        map(
            lambda node: (node, None)
            if isinstance(node, (Import, ImportFrom))
            else (None, node),
            parsed_ast.body,
        )
    )
    parsed_ast.body = list(
        filter(
            None,
            chain.from_iterable(
                (
                    parsed_ast.body[:1] if doc_str else iter(()),
                    sorted(
                        map(itemgetter(0), whole),
                        key=lambda import_from: getattr(import_from, "module", None)
                        == "__future__",
                        reverse=True,
                    ),
                    map(itemgetter(1), whole[1:] if doc_str else whole),
                ),
            ),
        )
    )
    return parsed_ast


def get_functions_and_classes(
    decorator_list,
    emit_call,
    emit_default_doc,
    emit_name,
    global__all__,
    input_mapping_it,
    name_tpl,
    no_word_wrap,
    parse_name,
):
    """
    :param decorator_list: List of decorators
    :type decorator_list: ```Optional[Union[List[Str], List[]]]```

    :param emit_call: Whether to emit a `__call__` method from the `_internal` IR subdict
    :type emit_call: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :param emit_name: Which type to generate.
    :type emit_name: ```Literal["argparse", "class", "function", "json_schema",
                                "pydantic", "sqlalchemy", "sqlalchemy_table"]```


    :param global__all__: `__all__` symbols for that magic
    :type global__all__: ```List[str]```

    :param input_mapping_it: Import location of mapping/2-tuple collection.
    :type input_mapping_it: ```Iterator[Tuple[str,AST]]```

    :param name_tpl: Template for the name, e.g., `{name}Config`.
    :type name_tpl: ```str```

    :param no_word_wrap: Whether word-wrap is disabled (on emission).
    :type no_word_wrap: ```Optional[Literal[True]]```

    :param parse_name: Which type to parse.
    :type parse_name: ```Literal["argparse", "class", "function", "json_schema",
                                 "pydantic", "sqlalchemy", "sqlalchemy_table"]```

    :return: Side-effect of appending `__all__`, this returns emitted values out of `input_mapping_it`
    :rtype: ```Tuple[AST]```
    """
    emitter = get_emitter(emit_name)
    return tuple(
        print("\nGenerating: {name!r}".format(name=name))
        or global__all__.append(name_tpl.format(name=name))
        or emitter(
            print("obj:", obj, ";") or get_parser(obj, parse_name)(obj),
            emit_default_doc=emit_default_doc,
            word_wrap=no_word_wrap is None,
            **get_emit_kwarg(decorator_list, emit_call, emit_name, name_tpl, name),
        )
        for name, obj in input_mapping_it
    )


def file_to_input_mapping(filepath, parse_name):
    """
    Create an `input_mapping` from a given file, i.e. Dict[str, AST]

    :param filepath: Location of JSON or Python file
    :type filepath: ```str```

    :param parse_name: Which type to parse.
    :type parse_name: ```Literal["argparse", "class", "function", "json_schema",
                                 "pydantic", "sqlalchemy", "sqlalchemy_table", "infer"]```

    :return: Dictionary of string (name) to AST node
    :rtype: ```dict``
    """
    if (
        parse_name == "json_schema"
        or parse_name == "infer"
        and filepath.endswith("{}json".format(path.extsep))
    ):
        with open(filepath, "rt") as f:
            json_contents = load(f)
        name = path.basename(filepath)
        if "name" not in json_contents:
            json_contents["name"] = pascal_to_upper_camelcase(name)
        input_mapping = {name: json_contents}
    else:
        with open(filepath, "rt") as f:
            mod = ast_parse(f.read())

        input_mapping = dict(
            map(
                lambda node: (node.name, node),
                filter(
                    rpartial(
                        isinstance,
                        tuple(kind2instance_type.values())
                        if parse_name == "infer"
                        else kind2instance_type[parse_name],
                    ),
                    mod.body,
                ),
            ),
        )
    return input_mapping


__all__ = [
    "file_to_input_mapping",
    "get_input_mapping_from_path",
    "get_emit_kwarg",
    "gen_file",
    "get_parser",
]



"""
Pydantic mocks
"""

from ast import AnnAssign, ClassDef, Index, Load, Name, Store, Subscript

from cdd.shared.ast_utils import maybe_type_comment, set_value

pydantic_class_str = """
class Cat(BaseModel):
    pet_type: Literal['cat']
    cat_name: str
"""

pydantic_class_cls_def = ClassDef(
    bases=[Name(ctx=Load(), id="BaseModel")],
    body=[
        AnnAssign(
            annotation=Subscript(
                ctx=Load(),
                slice=Index(value=set_value("cat")),
                value=Name(ctx=Load(), id="Literal"),
            ),
            simple=1,
            target=Name(ctx=Store(), id="pet_type"),
            value=None,
            expr=None,
            expr_target=None,
            expr_annotation=None,
            lineno=None,
            **maybe_type_comment
        ),
        AnnAssign(
            annotation=Name(ctx=Load(), id="str"),
            simple=1,
            target=Name(ctx=Store(), id="cat_name"),
            value=None,
            expr=None,
            expr_target=None,
            expr_annotation=None,
            lineno=None,
            **maybe_type_comment
        ),
    ],
    decorator_list=[],
    keywords=[],
    name="Cat",
)

__all__ = ["pydantic_class_str", "pydantic_class_cls_def"]



"""
FastAPI utils
"""

from functools import partial

from cdd.shared.ast_utils import Dict_to_dict, get_value


def model_handler(key, model_name, location, mime_type):
    """
    Create fully-qualified model name from unqualified name

    :param key: Key name
    :type key: ```str```

    :param model_name: Not fully-qualified model name or a `{"$ref": string}` dict
    :type model_name: ```str|dict```

    :param location: Full-qualified parent path
    :type location: ```str```

    :param mime_type: MIME type
    :type mime_type: ```str```

    :return: Tuple["content", JSON ref to model name, of form `{"$ref": string}`]
    :rtype: ```Tuple[Union[str,"content"], dict]```
    """
    return (
        (key, model_name)
        if isinstance(model_name, dict)
        else (
            "content",
            {mime_type: {"schema": {"$ref": "{}{}".format(location, model_name)}}},
        )
    )


parse_handlers = {
    "model": partial(
        model_handler, location="#/components/schemas/", mime_type="application/json"
    )
}


def parse_fastapi_responses(responses):
    """
    Parse FastAPI "responses" key

    :param responses: `responses` keyword value from FastAPI decorator on route
    :type responses: ```Dict```

    :return: Transformed FastAPI "responses"
    :rtype: ```dict```
    """

    return {
        key: dict(
            (
                (
                    lambda _v: (parse_handlers[k](k, _v))
                    if k in parse_handlers
                    else (k, _v)
                )(get_value(v))
            )
            for k, v in Dict_to_dict(val).items()
        )
        for key, val in Dict_to_dict(responses.value).items()
    }


__all__ = ["parse_fastapi_responses"]



"""
Helpers to traverse the AST, extract the docstring out, parse and format to intended style
"""

from ast import (
    AnnAssign,
    Assign,
    AsyncFunctionDef,
    ClassDef,
    FunctionDef,
    NodeTransformer,
    arg,
    get_docstring,
    walk,
)
from collections import OrderedDict
from copy import deepcopy
from operator import attrgetter, eq

import cdd.docstring.emit
import cdd.docstring.parse
import cdd.function.emit
import cdd.function.parse
import cdd.shared.ast_cst_utils
import cdd.shared.parse.utils.parser_utils
from cdd.shared.ast_utils import (
    annotate_ancestry,
    find_in_ast,
    maybe_type_comment,
    set_arg,
    set_docstring,
    set_value,
    to_annotation,
    to_type_comment,
)
from cdd.shared.cst_utils import reindent_block_with_pass_body
from cdd.shared.docstring_parsers import parse_docstring
from cdd.shared.pure_utils import PY_GTE_3_8, is_ir_empty, none_types, omit_whitespace
from cdd.shared.source_transformer import ast_parse


def has_type_annotations(node):
    """
    Whether the node—incl. any nodes within this node—have type annotations

    :param node: AST node
    :type node: ```AST```
    """
    return any(
        filter(
            lambda _node: hasattr(_node, "annotation")
            and _node.annotation is not None
            or hasattr(_node, "returns")
            and _node.returns is not None,
            walk(node),
        )
    )


class DocTrans(NodeTransformer):
    """
    Walk the nodes modifying the docstring and inlining||commenting types as it goes
    """

    def __init__(
        self,
        docstring_format,
        word_wrap,
        type_annotations,
        existing_type_annotations,
        whole_ast,
    ):
        """
        Transform the docstrings found to intended docstring_format, potentially manipulating type annotations also

        :param docstring_format: Format of docstring
        :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

        :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
        :type word_wrap: ```bool```

        :param type_annotations: True to have type annotations (3.6+), False to place in docstring
        :type type_annotations: ```bool```

        :param existing_type_annotations: Whether there are any type annotations (3.6+)
        :type existing_type_annotations: ```bool```

        :param whole_ast: The entire input AST, useful for lookups by location
        :type whole_ast: ```AST``
        """

        self.docstring_format = docstring_format
        self.word_wrap = word_wrap
        self.type_annotations = type_annotations
        self.existing_type_annotations = existing_type_annotations
        if not hasattr(whole_ast, "_location"):
            self.whole_ast = deepcopy(whole_ast)
            annotate_ancestry(self.whole_ast)
        else:
            self.whole_ast = whole_ast
        self.memoized = {}

    # def generic_visit(self, node):
    #     """
    #     visits the `AST` node, if it could have a docstring pass it off to that handler
    #
    #     :param node: The AST node
    #     :type node: ```AST```
    #
    #     :return: Potentially changed AST node
    #     :rtype: ```AST```
    #     """
    #     is_func, doc_str = isinstance(node, (AsyncFunctionDef, FunctionDef)), None
    #     if is_func or isinstance(node, ClassDef):
    #         node, doc_str = self._handle_node_with_docstring(node)
    #         doc_str = ast.get_docstring(node, clean=True)
    #     if is_func:
    #         node = self._handle_function(node, doc_str)
    #     return super(DocTrans, self).generic_visit(node)

    def visit_AnnAssign(self, node):
        """
        Handle `AnnAssign`

        :param node: AnnAssign
        :type node: ```AnnAssign```

        :return: `AnnAssign` if `type_annotations` and type found else `Assign`
        :rtype: ```Union[AnnAssign, Assign]```
        """
        if self.type_annotations:
            node.annotation = self._get_ass_typ(node)
            setattr(node, "type_comment", None)
            return node

        return Assign(
            targets=[node.target],
            lineno=node.lineno,
            col_offset=getattr(node, "col_offset", None),
            end_lineno=getattr(node, "end_lineno", None),
            end_col_offset=getattr(node, "end_col_offset", None),
            type_comment=to_type_comment(node.annotation),
            # `var: int` is valid and turning it to `var = None  # type_comment int` would
            # be wrong, as the upcoming smarter type tracer will reverse this to `var: Optional[int] = None`
            value=set_value(none_types[-1]) if node.value is None else node.value,
        )

    def visit_Assign(self, node):
        """
        Handle `Assign`

        :param node: Assign
        :type node: ```Assign```

        :return: `AnnAssign` if `type_annotations` and type found else `Assign`
        :rtype: ```Union[Assign, AnnAssign]```
        """
        typ = self._get_ass_typ(node)
        annotation = (
            None if not self.type_annotations or typ is None else to_annotation(typ)
        )
        if annotation:
            assert len(node.targets) == 1
            return AnnAssign(
                annotation=to_annotation(typ),
                lineno=node.lineno,
                col_offset=getattr(node, "col_offset", None),
                end_lineno=getattr(node, "end_lineno", None),
                end_col_offset=getattr(node, "end_col_offset", None),
                simple=1,
                target=node.targets[0],
                expr=None,
                expr_target=None,
                expr_annotation=None,
                value=node.value,
                **maybe_type_comment,
            )
        else:
            setattr(node, "type_comment", typ)
        return node

    def visit_FunctionDef(self, node):
        """
        visits the `FunctionDef`, potentially augmenting its docstring and argument types

        :param node: FunctionDef
        :type node: ```FunctionDef```

        :return: Potentially changed FunctionDef
        :rtype: ```FunctionDef```
        """
        return self._handle_function(node, get_docstring(node, clean=False))

    def _get_ass_typ(self, node):
        """
        Get the type of the assignment

        :param node: Assignment
        :type node: ```Union[Assign, AnnAssign]```

        :return: The type of the assignment, e.g., `int`
        :rtype: ```Optional[str]```
        """
        name, typ_dict = (
            (node.targets[0], {"typ": getattr(node, "type_comment", None)})
            if isinstance(node, Assign)
            else (
                node.target,
                {"typ": node.annotation or getattr(node, "type_comment", None)},
            )
        )
        if not hasattr(node, "_location") or node._location[:-1] == [None]:
            return typ_dict["typ"]
        search = node._location[:-1]
        search_str = ".".join(search)

        self.memoized[search_str] = ir = self.memoized.get(
            search_str,
            (
                lambda parent: (
                    (
                        lambda doc_str: None
                        if doc_str is None
                        else cdd.docstring.parse.docstring(doc_str)
                    )(get_docstring(parent, clean=False))
                    if isinstance(parent, (ClassDef, AsyncFunctionDef, FunctionDef))
                    else {"params": OrderedDict()}
                )
            )(find_in_ast(search, self.whole_ast)),
        ) or {"params": OrderedDict()}

        return ir["params"].get(name, typ_dict)[
            "typ"
        ]  # if ir is not None and ir.get("params") else None

    def _handle_function(self, node, original_doc_str):
        """
        Handle functions

        :param node: AsyncFunctionDef | FunctionDef
        :type node: ```Union[AsyncFunctionDef, FunctionDef]```

        :param original_doc_str: The docstring
        :type original_doc_str: ```Optional[str]```

        :return: Same type as input with args, returns, and docstring potentially modified
        :rtype: ```Union[AsyncFunctionDef, FunctionDef]```
        """
        ir = parse_docstring(original_doc_str)
        cdd.shared.parse.utils.parser_utils.ir_merge(
            ir, cdd.function.parse.function(node)
        )
        ir["name"] = node.name
        indent_level = max(
            len(node._location), 1
        )  # function docstrings always have at least 1 indent level
        doc_str = (
            None
            if is_ir_empty(ir)
            else cdd.docstring.emit.docstring(
                ir,
                emit_types=not self.type_annotations,
                emit_default_doc=False,
                docstring_format=self.docstring_format,
                indent_level=indent_level,
                word_wrap=self.word_wrap,
            )
        )
        if not doc_str or doc_str.isspace():
            if original_doc_str is not None:
                del node.body[0]
        else:
            set_docstring(
                original_doc_str
                if original_doc_str
                and eq(*map(omit_whitespace, (original_doc_str, doc_str)))
                else doc_str,
                False,
                node,
            )
        if self.type_annotations:
            # Add annotations
            if ir["params"]:
                node.args.args = list(
                    map(
                        lambda _arg: arg(
                            arg=_arg.arg,
                            annotation=to_annotation(
                                _arg.annotation
                                if _arg.annotation is not None
                                or _arg.arg in frozenset(("self", "cls"))
                                else ir["params"][_arg.arg].get("typ")
                            ),
                            identifier_arg=None,
                            end_col_offset=getattr(_arg, "end_col_offset", None),
                            **dict(expr=None, **maybe_type_comment)
                            if PY_GTE_3_8
                            else {},
                        ),
                        node.args.args,
                    )
                )
            if (
                "return_type" in (ir.get("returns") or iter(()))
                and ir["returns"]["return_type"].get("typ") is not None
            ):
                node.returns = to_annotation(ir["returns"]["return_type"]["typ"])
        else:
            # Remove annotations
            node.args.args = list(map(set_arg, map(attrgetter("arg"), node.args.args)))
            node.returns = None

        node.body = list(map(self.visit, node.body))
        return node


def clear_annotation(node):
    """
    Remove annotations and type_comments from node

    :param node: AST node
    :type node: ```AST```

    :return: AST node with annotations and type_comments set to `None`
    :rtype: ```AST```
    """
    if getattr(node, "annotation", None) is not None:
        node.annotation = None
    if getattr(node, "type_comment", None) is not None:
        setattr(node, "type_comment", None)
    return node


def doctransify_cst(cst_list, node):
    """
    Carefully replace only docstrings, function return annotations, assignment and annotation assignments.
    (maintaining all other existing whitespace, comments, &etc.); and only when cdd has changed them

    :param cst_list: List of `namedtuple`s with at least ("line_no_start", "line_no_end", "value") attributes
    :type cst_list: ```List[NamedTuple]```

    :param node: AST node with a `.body`, probably the `ast.Module`
    :type node: ```AST```
    """
    for _node in walk(node):
        if hasattr(_node, "_location"):
            is_func = isinstance(_node, (AsyncFunctionDef, FunctionDef))
            if isinstance(_node, ClassDef) or is_func:
                cst_idx, cst_node = cdd.shared.ast_cst_utils.find_cst_at_ast(
                    cst_list, _node
                )

                if cst_node is not None:
                    cdd.shared.ast_cst_utils.maybe_replace_doc_str_in_function_or_class(
                        _node, cst_idx, cst_list
                    )

                    if is_func:
                        cur_ast_node = ast_parse(
                            reindent_block_with_pass_body(cst_list[cst_idx].value),
                            skip_annotate=True,
                            skip_docstring_remit=True,
                        ).body[0]

                        cdd.shared.ast_cst_utils.maybe_replace_function_return_type(
                            _node, cur_ast_node, cst_idx, cst_list
                        )
                        cdd.shared.ast_cst_utils.maybe_replace_function_args(
                            _node, cur_ast_node, cst_idx, cst_list
                        )
            # TODO: AnnAssign|Assign
            # AnnAssign|Assign is separate task than the `maybe_replace_function_args` as inferring types is done
            #   better with knowledge of function return types and function arguments (`default` being the only issue)
            # elif isinstance(_node, (AnnAssign, Assign)):
            #     print("(AnnAssign | Assign)._location:", _node._location, ";")
            #     print_ast(_node)


__all__ = ["DocTrans", "clear_annotation", "doctransify_cst", "has_type_annotations"]



"""
Tests for `cdd.emit.sqlalchemy`
"""

import os
from copy import deepcopy
from platform import system
from unittest import TestCase, skipIf

import cdd.argparse_function.emit
import cdd.argparse_function.parse
import cdd.class_.emit
import cdd.docstring.emit
import cdd.function.emit
import cdd.json_schema.emit
import cdd.shared.emit.file
import cdd.sqlalchemy.emit
from cdd.tests.mocks.ir import (
    intermediate_repr_empty,
    intermediate_repr_no_default_sql_doc,
)
from cdd.tests.mocks.sqlalchemy import (
    config_decl_base_ast,
    config_tbl_with_comments_ast,
    empty_with_inferred_pk_column_assign,
)
from cdd.tests.utils_for_tests import run_ast_test, unittest_main


class TestEmitSqlAlchemy(TestCase):
    """Tests emission"""

    def test_to_sqlalchemy_table(self) -> None:
        """
        Tests that `emit.sqlalchemy_table` with `intermediate_repr_no_default_sql_doc` produces `config_tbl_ast`
        """
        gen_ast = cdd.sqlalchemy.emit.sqlalchemy_table(
            deepcopy(intermediate_repr_no_default_sql_doc), name="config_tbl"
        )
        run_ast_test(
            self,
            gen_ast=gen_ast,
            gold=config_tbl_with_comments_ast,
        )

    def test_to_sqlalchemy_table_with_inferred_pk(self) -> None:
        """
        Tests that `emit.sqlalchemy_table` with `intermediate_repr_no_default_sql_doc` produces `config_tbl_ast`
        """
        sql_table = cdd.sqlalchemy.emit.sqlalchemy_table(
            deepcopy(intermediate_repr_empty), name="empty_with_inferred_pk_tbl"
        )
        run_ast_test(
            self,
            sql_table,
            gold=empty_with_inferred_pk_column_assign,
        )

    @skipIf(
        "GITHUB_ACTIONS" in os.environ and system() in frozenset(("Darwin", "Linux")),
        "GitHub Actions fails this test on macOS & Linux (unable to replicate locally)",
    )
    def test_to_sqlalchemy(self) -> None:
        """
        Tests that `emit.sqlalchemy` with `intermediate_repr_no_default_sql_doc` produces `config_tbl_ast`
        """
        system() in frozenset(("Darwin", "Linux")) and print("test_to_sqlalchemy")

        ir = deepcopy(intermediate_repr_no_default_sql_doc)
        ir["name"] = "Config"
        gen_ast = cdd.sqlalchemy.emit.sqlalchemy(
            ir,
            # class_name="Config",
            table_name="config_tbl",
        )
        run_ast_test(
            self,
            gen_ast=gen_ast,
            gold=config_decl_base_ast,
        )


unittest_main()



"""
Tests for `cdd.emit.utils.json_schema_utils`
"""

from ast import Dict, List, Load, Set, Tuple
from copy import deepcopy
from typing import List as TList
from unittest import TestCase

from cdd.json_schema.utils.emit_utils import param2json_schema_property
from cdd.shared.ast_utils import set_value
from cdd.tests.mocks.ir import intermediate_repr_no_default_doc
from cdd.tests.utils_for_tests import unittest_main


class TestEmitJsonSchemaUtils(TestCase):
    """Tests cdd.emit.utils.json_schema_utils"""

    required: TList[str] = []
    name = "param"
    default_ = tuple(range(5))
    default_elts = list(map(set_value, default_))

    def test_param2json_schema_property_default_property(self) -> None:
        """
        Tests that `param2json_schema_property` with different default types
        """

        self.assertEqual(
            param2json_schema_property(
                (
                    self.name,
                    {
                        "default": Set(
                            elts=self.default_elts,
                            ctx=Load(),
                            expr=None,
                        )
                    },
                ),
                required=self.required,
            ),
            (self.name, {"default": set(self.default_)}),
        )

        self.assertEqual(
            param2json_schema_property(
                (
                    self.name,
                    {"default": Tuple(elts=self.default_elts, ctx=Load(), expr=None)},
                ),
                required=self.required,
            ),
            (self.name, {"default": self.default_}),
        )

        self.assertEqual(
            param2json_schema_property(
                (
                    self.name,
                    {"default": List(elts=self.default_elts, ctx=Load(), expr=None)},
                ),
                required=self.required,
            ),
            (self.name, {"default": list(self.default_)}),
        )

        self.assertEqual(
            param2json_schema_property(
                (
                    self.name,
                    {"default": None},
                ),
                required=self.required,
            ),
            (self.name, {}),
        )

        self.assertEqual(
            param2json_schema_property(
                (
                    self.name,
                    {
                        "default": Dict(
                            keys=[set_value("Foo")],
                            values=[set_value("bar")],
                            expr=None,
                        )
                    },
                ),
                required=self.required,
            ),
            (self.name, {"default": {"Foo": "bar"}}),
        )

    def test_param2json_schema_property_choices(self) -> None:
        """
        Tests that `param2json_schema_property` with different choices as `Set`
        """
        self.assertEqual(
            param2json_schema_property(
                (
                    self.name,
                    {
                        "choices": Set(
                            elts=self.default_elts,
                            ctx=Load(),
                            expr=None,
                        )
                    },
                ),
                required=self.required,
            ),
            (self.name, {"pattern": "0|1|2|3|4"}),
        )

    def test_param2json_schema_property_type_datetime(self) -> None:
        """
        Tests that `param2json_schema_property` with different type as `datetime`
        """
        self.assertEqual(
            param2json_schema_property(
                (
                    self.name,
                    {"typ": "datetime"},
                ),
                required=self.required,
            ),
            (self.name, {"type": "string", "format": "date-time"}),
        )

    def test_param2json_schema_property_type_boolean(self) -> None:
        """
        Tests that `param2json_schema_property` with different type as `datetime`
        """
        ir = deepcopy(intermediate_repr_no_default_doc)
        self.assertEqual(
            param2json_schema_property(
                (
                    self.name,
                    {"typ": "bool"},
                ),
                required=self.required,
            ),
            (self.name, {"type": "boolean"}),
        )

        required: TList[str] = []
        self.assertEqual(
            param2json_schema_property(
                (
                    self.name,
                    ir["params"]["as_numpy"],
                ),
                required=required,
            ),
            (
                self.name,
                {
                    "description": ir["params"]["as_numpy"]["description"],
                    "type": "boolean",
                },
            ),
        )
        self.assertListEqual(required, [])


unittest_main()



"""
Tests for setup.py
"""

from importlib.machinery import SourceFileLoader
from importlib.util import module_from_spec, spec_from_loader
from operator import methodcaller
from os import path
from os.path import extsep
from sys import modules
from unittest import TestCase
from unittest.mock import patch

from cdd.tests.utils_for_tests import mock_function, unittest_main


class TestSetupPy(TestCase):
    """
    Tests whether docstrings are parsed out—and emitted—correctly
    """

    @classmethod
    def setUpClass(cls) -> None:
        """Construct the setup_py module"""
        cls.mod = cls.import_setup_py()

    @staticmethod
    def import_setup_py():
        """
        Function which imports setup.py as a module

        :return: setup.py as a module
        :rtype: ```Union[module, ModuleSpec]```
        """
        modname = "setup_py"
        loader = SourceFileLoader(
            modname,
            path.join(
                path.dirname(path.dirname(path.dirname(__file__))),
                "setup{extsep}py".format(extsep=extsep),
            ),
        )
        modules[modname] = module_from_spec(spec_from_loader(loader.name, loader))
        loader.exec_module(modules[modname])
        return modules[modname]

    def test_properties(self) -> None:
        """
        Tests whether 'setup.py' has correct properties
        """
        self.assertEqual(getattr(self.mod, "package_name"), "cdd")
        self.assertEqual(self.mod.__name__, "setup_py")

    def test_to_funcs(self) -> None:
        """Tests that `to_funcs` produces the right local and install dirs"""
        to_funcs = getattr(self.mod, "to_funcs")
        args = "5", "6"
        local_dir_join_func_resp, install_dir_join_func_resp = map(
            methodcaller("__call__"), to_funcs(*args)
        )
        self.assertNotEqual(local_dir_join_func_resp, install_dir_join_func_resp)
        self.assertEqual(
            local_dir_join_func_resp,
            path.join(path.dirname(path.dirname(__file__)), *args),
        )

    def test_main(self) -> None:
        """
        Tests that no errors occur in `main` function call (up to `setup()`, which is tested in setuptools)
        """
        with patch("setup_py.setup", mock_function):
            self.assertIsNone(self.mod.main())

    def test_setup_py_main(self) -> None:
        """
        Tests that `__name__ == __main__` calls the `main` function via `setup_py_main` call
        """

        with patch("setup_py.main", mock_function), patch(
            "setup_py.__name__", "__main__"
        ):
            self.assertIsNone(self.mod.setup_py_main())


unittest_main()



"""
Not a dead module
"""

from ast import Assign, Expr, ImportFrom, List, Load, Module, Name, Store, alias
from collections import OrderedDict, deque
from functools import partial
from importlib import import_module
from inspect import getfile
from itertools import chain, groupby
from operator import itemgetter
from os import makedirs, path

import cdd.class_.parse
import cdd.shared.emit.file
from cdd.compound.exmod_utils import emit_file_on_hierarchy, get_module_contents
from cdd.shared.ast_utils import maybe_type_comment, set_value
from cdd.shared.pkg_utils import relative_filename
from cdd.shared.pure_utils import INIT_FILENAME


def exmod(
    emit_name,
    module,
    blacklist,
    whitelist,
    output_directory,
    mock_imports,
    no_word_wrap,
    dry_run,
    filesystem_layout="as_input",
):
    """
    Expose module as `emit` types into `output_directory`

    :param emit_name: What type(s) to generate.
    :type emit_name: ```List[Literal["argparse", "class", "function", "json_schema",
                                     "pydantic", "sqlalchemy", "sqlalchemy_table"]]```

    :param module: Module name or path
    :type module: ```str```

    :param blacklist: Modules/FQN to omit. If unspecified will emit all (unless whitelist).
    :type blacklist: ```Union[List[str],Tuple[str]]```

    :param whitelist: Modules/FQN to emit. If unspecified will emit all (minus blacklist).
    :type whitelist: ```Union[List[str],Tuple[str]]```

    :param output_directory: Where to place the generated exposed interfaces to the given `--module`.
    :type output_directory: ```str```

    :param mock_imports: Whether to generate mock TensorFlow imports
    :type mock_imports: ```bool```

    :param no_word_wrap: Whether word-wrap is disabled (on emission).
    :type no_word_wrap: ```Optional[Literal[True]]```

    :param dry_run: Show what would be created; don't actually write to the filesystem
    :type dry_run: ```bool```

    :param filesystem_layout: Hierarchy of folder and file names generated. "java" is file per package per name.
    :type filesystem_layout: ```Literal["java", "as_input"]```
    """
    if not isinstance(emit_name, str):
        deque(
            map(
                partial(
                    exmod,
                    module=module,
                    blacklist=blacklist,
                    whitelist=whitelist,
                    mock_imports=mock_imports,
                    filesystem_layout=filesystem_layout,
                    no_word_wrap=no_word_wrap,
                    output_directory=output_directory,
                    dry_run=dry_run,
                ),
                emit_name or iter(()),
            ),
            maxlen=0,
        )
    elif dry_run:
        print("mkdir\t{output_directory!r}".format(output_directory=output_directory))
    elif not path.isdir(output_directory):
        makedirs(output_directory)

    emit_name = (
        emit_name[0]
        if emit_name is not None and len(emit_name) == 1 and isinstance(emit_name, list)
        else emit_name
    )
    assert isinstance(
        emit_name, (str, type(None))
    ), "Expected `str` got `{emit_name_type!r}`".format(emit_name_type=type(emit_name))

    module_name, new_module_name = map(path.basename, (module, output_directory))
    module = (
        # partial(module_from_file, module_name=module_name)
        # if path.isdir(module) else
        import_module
    )(module)

    # if module.__file__ is None:
    #     raise ModuleNotFoundError(module_name)

    module_root_dir = path.dirname(module.__file__) + path.sep

    mod_path = ".".join((path.basename(module_root_dir[:-1]), module_name))
    blacklist, whitelist = map(
        frozenset, (blacklist or iter(()), whitelist or iter(()))
    )
    proceed = any(
        (
            sum(map(len, (blacklist, whitelist))) == 0,
            mod_path not in blacklist and (mod_path in whitelist or not whitelist),
        )
    )
    if not proceed:
        return

    _emit_file_on_hierarchy = partial(
        emit_file_on_hierarchy,
        emit_name=emit_name,
        module_name=module_name,
        new_module_name=new_module_name,
        mock_imports=mock_imports,
        filesystem_layout=filesystem_layout,
        output_directory=output_directory,
        no_word_wrap=no_word_wrap,
        dry_run=dry_run,
    )

    # Might need some `groupby` in case multiple files are in the one project; same for `get_module_contents`

    imports = list(
        map(
            _emit_file_on_hierarchy,
            map(
                lambda name_source: (
                    name_source[0],
                    (
                        lambda filename: filename[len(module_name) + 1 :]
                        if filename.startswith(module_name)
                        else filename
                    )(relative_filename(getfile(name_source[1]))),
                    {"params": OrderedDict(), "returns": OrderedDict()}
                    if dry_run
                    else cdd.class_.parse.class_(
                        name_source[1], merge_inner_function="__init__"
                    ),
                ),
                map(
                    lambda name_source: (
                        name_source[0][len(module_name) + 1 :],
                        name_source[1],
                    ),
                    get_module_contents(
                        module, module_root_dir=module_root_dir
                    ).items(),
                ),
            ),
        ),
    )
    assert len(imports), "Module contents are empty"
    modules_names = tuple(
        map(
            lambda name_module: (
                name_module[0],
                tuple(map(itemgetter(1), name_module[1])),
            ),
            groupby(
                map(
                    lambda node_mod: (
                        node_mod[0],
                        node_mod[2].module,
                    ),
                    imports,
                ),
                itemgetter(0),
            ),
        )
    )

    init_filepath = path.join(output_directory, new_module_name, INIT_FILENAME)
    if dry_run:
        print("write\t{init_filepath!r}".format(init_filepath=init_filepath))
    else:
        cdd.shared.emit.file.file(
            Module(
                body=list(
                    chain.from_iterable(
                        (
                            (Expr(set_value("\nExport internal imports\n")),),
                            map(
                                lambda module_names: ImportFrom(
                                    module=module_names[0],
                                    names=list(
                                        map(
                                            lambda names: alias(
                                                names,
                                                None,
                                                identifier=None,
                                                identifier_name=None,
                                            ),
                                            module_names[1],
                                        )
                                    ),
                                    level=1,
                                    identifier=None,
                                ),
                                modules_names,
                            ),
                            (
                                Assign(
                                    targets=[Name("__all__", Store())],
                                    value=List(
                                        ctx=Load(),
                                        elts=list(
                                            map(
                                                set_value,
                                                sorted(
                                                    frozenset(
                                                        chain.from_iterable(
                                                            map(
                                                                itemgetter(1),
                                                                modules_names,
                                                            )
                                                        ),
                                                    )
                                                ),
                                            )
                                        ),
                                        expr=None,
                                    ),
                                    expr=None,
                                    lineno=None,
                                    **maybe_type_comment
                                ),
                            ),
                        )
                    )
                ),
                stmt=None,
                type_ignores=[],
            ),
            init_filepath,
            mode="wt",
        )


__all__ = ["exmod"]



"""
Mocks for JSON Schema
"""

from copy import deepcopy

from cdd.tests.mocks.docstrings import docstring_header_and_return_no_nl_str

config_schema = {
    "$id": "https://offscale.io/config.schema.json",
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "description": docstring_header_and_return_no_nl_str,
    "type": "object",
    "properties": {
        "dataset_name": {
            "description": "[PK] name of dataset.",
            "type": "string",
            "default": "mnist",
        },
        "tfds_dir": {
            "description": "directory to look for models in.",
            "type": "string",
            "default": "~/tensorflow_datasets",
        },
        "K": {
            "description": "backend engine, e.g., `np` or `tf`.",
            "type": "string",
            "pattern": "np|tf",
            "default": "np",
        },
        "as_numpy": {
            "description": "Convert to numpy ndarrays",
            "type": "boolean",
        },
        "data_loader_kwargs": {
            "description": "pass this as arguments to data_loader function",
            "type": "object",
        },
    },
    "required": ["dataset_name", "tfds_dir", "K"],
}

config_schema_with_sql_types = deepcopy(config_schema)
for param, typ in (
    ("dataset_name", "String"),
    ("tfds_dir", "String"),
    ("as_numpy", "Boolean"),
    ("data_loader_kwargs", "JSON"),
):
    config_schema_with_sql_types["properties"][param]["x_typ"] = {"sql": {"type": typ}}


server_error_schema = {
    "$id": "https://offscale.io/error_json.schema.json",
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "description": "Error schema",
    "type": "object",
    "properties": {
        "error": {"description": "Name of the error", "type": "string"},
        "error_description": {
            "description": "Description of the error",
            "type": "string",
        },
        "error_code": {
            "description": "Code of the error (usually is searchable in a KB for further information)",
            "type": "string",
        },
        "status_code": {
            "description": "Status code (usually for HTTP)",
            "type": "number",
        },
    },
    "required": ["error", "error_description"],
}


__all__ = ["config_schema", "server_error_schema", "config_schema_with_sql_types"]



"""
Tests for source_transformer
"""

from ast import FunctionDef, Pass, arguments
from unittest import TestCase

from cdd.shared.pure_utils import tab
from cdd.shared.source_transformer import to_code
from cdd.tests.utils_for_tests import unittest_main


class TestSourceTransformer(TestCase):
    """
    Tests for source_transformer
    """

    def test_to_code(self) -> None:
        """
        Tests to_source in Python 3.9 and < 3.9
        """
        func_def = FunctionDef(
            name="funcy",
            args=arguments(
                posonlyargs=[],
                args=[],
                kwonlyargs=[],
                kw_defaults=[],
                defaults=[],
                vararg=None,
                kwarg=None,
            ),
            body=[Pass()],
            decorator_list=[],
            lineno=None,
            returns=None,
        )

        self.assertEqual(
            to_code(func_def).rstrip("\n"),
            "def funcy():\n" "{tab}pass".format(tab=tab),
        )


unittest_main()



"""
Tests for the Intermediate Representation produced by the `class` and `function` parsers
"""

import ast
from collections import OrderedDict
from operator import itemgetter
from unittest import TestCase
from unittest.mock import patch

import cdd.argparse_function.emit
import cdd.class_.parse
import cdd.json_schema.parse
from cdd.shared.ast_utils import RewriteAtQuery
from cdd.tests.mocks.classes import (
    class_ast,
    class_google_tf_tensorboard_ast,
    class_google_tf_tensorboard_str,
    class_reduction_v2,
    class_torch_nn_l1loss_ast,
    class_torch_nn_l1loss_str,
    class_torch_nn_one_cycle_lr_ast,
    class_torch_nn_one_cycle_lr_str,
)
from cdd.tests.mocks.ir import (
    class_google_tf_tensorboard_ir,
    class_torch_nn_l1loss_ir,
    class_torch_nn_one_cycle_lr_ir,
    docstring_google_tf_adadelta_function_ir,
    intermediate_repr_no_default_doc,
)
from cdd.tests.mocks.methods import docstring_google_tf_adadelta_function_str
from cdd.tests.utils_for_tests import inspectable_compile, run_ast_test, unittest_main


class TestParseClass(TestCase):
    """
    Tests whether the intermediate representation is consistent when parsed from different inputs.

    IR is a dictionary of form:
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    """

    def test_from_class(self) -> None:
        """
        Tests whether `class_` produces `intermediate_repr_no_default_doc`
              from `class_ast`
        """
        ir = cdd.class_.parse.class_(class_ast)
        del ir["_internal"]  # Not needed for this test
        ir["name"] = None
        self.assertDictEqual(ir, intermediate_repr_no_default_doc)

    def test_from_class_in_memory(self) -> None:
        """
        Tests that `parse.class_` produces properly from a `class` in memory of current interpreter
        """

        class A(object):
            """A is one boring class"""

        ir = cdd.class_.parse.class_(A)
        del ir["_internal"]  # Not needed for this test
        self.assertDictEqual(
            ir,
            {
                "doc": "A is one boring class",
                "name": "TestParseClass.test_from_class_in_memory.<locals>.A",
                "params": OrderedDict(),
                "returns": None,
            },
        )

    def test_from_adadelta_class_in_memory(self) -> None:
        """
        Tests that parse.class produces properly from a `class` in memory of current interpreter
        """
        Adadelta = getattr(
            inspectable_compile(docstring_google_tf_adadelta_function_str),
            "Adadelta",
        )
        ir = cdd.class_.parse.class_(Adadelta)
        del ir["_internal"]
        # self.assertDictEqual(ir, docstring_google_tf_adadelta_ir)
        self.assertDictEqual(
            ir,
            docstring_google_tf_adadelta_function_ir,
        )

    def test_from_class_and_function(self) -> None:
        """
        Tests that the parser can combine the outer class docstring + structure
        with the inner function parameter defaults
        """

        # Sanity check
        run_ast_test(
            self,
            class_google_tf_tensorboard_ast,
            gold=ast.parse(class_google_tf_tensorboard_str).body[0],
        )

        parsed_ir = cdd.class_.parse.class_(
            class_google_tf_tensorboard_ast,
            merge_inner_function="__init__",
            parse_original_whitespace=True,
            word_wrap=False,
            infer_type=True,
        )

        del parsed_ir["_internal"]  # Not needed for this test

        self.assertEqual(
            *map(itemgetter("doc"), (parsed_ir, class_google_tf_tensorboard_ir))
        )
        self.assertDictEqual(parsed_ir, class_google_tf_tensorboard_ir)

    def test_from_class_and_function_in_memory(self) -> None:
        """
        Tests that the parser can combine the outer class docstring + structure
        with the inner function parameter defaults
        """

        parsed_ir = cdd.class_.parse.class_(
            RewriteAtQuery,
            merge_inner_function="__init__",
            infer_type=True,
        )

        del parsed_ir["_internal"]  # Not needed for this test

        self.assertDictEqual(
            parsed_ir,
            {
                "doc": "Replace the node at query with given node",
                "name": "RewriteAtQuery",
                "params": OrderedDict(
                    (
                        (
                            "search",
                            {
                                "doc": "Search query, e.g., "
                                "['node_name', 'function_name', "
                                "'arg_name']",
                                "typ": "List[str]",
                            },
                        ),
                        (
                            "replacement_node",
                            {"doc": "Node to replace this search", "typ": "AST"},
                        ),
                        (
                            "replaced",
                            {
                                "doc": "Whether a node has been replaced "
                                "(only replaces first "
                                "occurrence)"
                            },
                        ),
                    )
                ),
                "returns": None,
            },
        )

    def test_from_class_and_function_torch(self) -> None:
        """
        Tests that the parser can combine the outer class docstring + structure
        with the inner function parameter defaults, given a PyTorch loss class
        """

        # Sanity check
        run_ast_test(
            self,
            class_torch_nn_l1loss_ast,
            gold=ast.parse(class_torch_nn_l1loss_str).body[0],
        )

        parsed_ir = cdd.class_.parse.class_(
            class_torch_nn_l1loss_ast,
            merge_inner_function="__init__",
            infer_type=True,
            parse_original_whitespace=True,
        )

        del parsed_ir["_internal"]  # Not needed for this test

        self.assertDictEqual(parsed_ir, class_torch_nn_l1loss_ir)

    def test_from_class_torch_nn_one_cycle_lr(self) -> None:
        """
        Tests that the parser can combine the outer class docstring + structure
        with the inner function parameter defaults, given a PyTorch loss LR scheduler class
        """

        # Sanity check
        run_ast_test(
            self,
            class_torch_nn_one_cycle_lr_ast,
            gold=ast.parse(class_torch_nn_one_cycle_lr_str).body[0],
        )

        parsed_ir = cdd.class_.parse.class_(
            class_torch_nn_one_cycle_lr_ast,
            merge_inner_function="__init__",
            infer_type=True,
            word_wrap=False,
        )

        del parsed_ir["_internal"]  # Not needed for this test

        self.assertEqual(parsed_ir["params"]["last_epoch"]["typ"], "int")
        self.assertEqual(parsed_ir["doc"], class_torch_nn_one_cycle_lr_ir["doc"])
        self.assertDictEqual(parsed_ir, class_torch_nn_one_cycle_lr_ir)

    def test__class_from_memory(self) -> None:
        """
        Tests that the parser can combine the outer class docstring + structure
        with the inner function parameter defaults, given a PyTorch loss LR scheduler class
        """

        class A(object):
            """A is one boring class"""

        with patch("inspect.getsourcefile", lambda _: None):
            ir = cdd.class_.parse._class_from_memory(
                A, A.__name__, False, False, False, False
            )
        self.assertDictEqual(
            ir,
            {
                "doc": "A is one boring class",
                "name": "A",
                "params": OrderedDict(),
                "returns": None,
            },
        )

    def test_from_class_class_reduction_v2(self) -> None:
        """
        Test class_reduction_v2 produces correct IR
        """
        ir = cdd.class_.parse.class_(class_reduction_v2)
        self.assertEqual(
            ir["params"],
            OrderedDict(
                (
                    ("AUTO", {"default": "auto", "typ": "str"}),
                    ("NONE", {"default": "none", "typ": "str"}),
                    ("SUM", {"default": "sum", "typ": "str"}),
                    (
                        "SUM_OVER_BATCH_SIZE",
                        {"default": "sum_over_batch_size", "typ": "str"},
                    ),
                )
            ),
        )
        self.assertEqual(ir["returns"], None)


unittest_main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Root __init__
"""

import logging
from logging import getLogger as get_logger

__author__ = "Samuel Marks"
__version__ = "0.0.99-rc4"
__description__ = (
    "Open API to/fro routes, models, and tests. "
    "Convert between docstrings, classes, methods, argparse, pydantic, and SQLalchemy."
)


root_logger = get_logger()
logging.getLogger("blib2to3").setLevel(logging.WARNING)

__all__ = ["get_logger", "root_logger", "__description__", "__version__"]



# pragma: no cover
# flake8: noqa
# === Copyright 2022 under CC0
"""Module docstring goes here"""

from operator import add


class C(object):
    """My cls"""

    @staticmethod
    def add1(foo):
        """
        :param foo: a foo
        :type foo: ```int```

        :return: foo + 1
        :rtype: ```int```
        """

        """foo"""
        def g(): """foo : bar ; can"""; pass

        def h(): # stuff
            pass

        def adder(a: int,
                  b: int) -> int:
            """
            :param a: First arg

            :param b: Second arg

            :return: first + second arg
            """
            # fmt: off
            res: \
                int \
                = a + b
            return res

        r = (
            add(foo, 1)
            or
            adder(foo, 1)
        )
        if r:
            None
        elif r:
            True
            False
            # ([5,5] @ [5,5]) *\
            -5 / 7 ** 6 + \
            6.0 - 6e1 & 1+2.34j
            r <<= 5
            print(r)
        else:
            pass
        # fmt: on
        # That^ incremented `foo` by 1
        return r


# from contextlib import ContextDecorator

# with ContextDecorator():
#    pass


def f():
    return 1



"""
JSON schema parser
"""

from collections import OrderedDict
from copy import deepcopy
from functools import partial

from cdd.docstring.parse import docstring
from cdd.json_schema.utils.parse_utils import json_schema_property_to_param


def json_schema(json_schema_dict, parse_original_whitespace=False):
    """
    Parse a JSON schema into the IR

    :param json_schema_dict: A valid JSON schema as a Python dict
    :type json_schema_dict: ```dict```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :return: IR representation of the given JSON schema
    :rtype: ```dict```
    """
    # I suppose a JSON-schema validation routine could be executed here
    schema = deepcopy(json_schema_dict)

    required = frozenset(schema["required"]) if schema.get("required") else frozenset()
    _json_schema_property_to_param = partial(
        json_schema_property_to_param, required=required
    )

    ir = docstring(
        json_schema_dict.get("description", ""),
        emit_default_doc=False,
        parse_original_whitespace=parse_original_whitespace,
    )
    ir.update(
        {
            "params": (
                OrderedDict(
                    map(_json_schema_property_to_param, schema["properties"].items())
                )
                if "properties" in schema
                else OrderedDict()
            ),
            "name": json_schema_dict.get(
                "name",
                json_schema_dict.get("id", json_schema_dict.get("title", ir["name"])),
            ),
        }
    )
    return ir


__all__ = ["json_schema"]



"""
Functions which produce docstring portions from various inputs
"""

from collections import namedtuple
from enum import Enum
from functools import partial
from itertools import chain, takewhile
from operator import contains, eq, itemgetter, ne
from textwrap import indent

from cdd.shared.defaults_utils import set_default_doc
from cdd.shared.pure_utils import (
    count_chars_from,
    count_iter_items,
    fill,
    identity,
    indent_all_but_first,
    num_of_nls,
    omit_whitespace,
    rpartial,
    tab,
)


def emit_param_str(
    param,
    style,
    purpose,
    emit_doc=True,
    emit_type=True,
    word_wrap=True,
    emit_default_doc=True,
):
    """
    Produce the docstring param/return lines

    :param param: Name, dict with keys: 'typ', 'doc', 'default'
    :type param: ```Tuple[str, dict]```

    :param style: the style of docstring
    :type style: ```Literal['rest', 'numpydoc', 'google']```

    :param purpose: Emit `:param` if purpose == 'function' elif purpose == 'class' then `:cvar` (ReST only)
    :type purpose: ```Literal['class', 'function']```

    :param emit_doc: Whether to emit the doc
    :type emit_doc: ```bool```

    :param emit_type: Whether to emit the type
    :type emit_type: ```bool```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :return: Newline joined pair of param, type
    :rtype: ```str```
    """
    name, _param = param
    del param

    _fill = fill if word_wrap else identity

    if style == "rest":
        emit_type &= purpose == "function"
        key, key_typ = (
            ("return", "rtype")
            if name == "return_type" and purpose != "class"
            else (
                "{var} {name}".format(
                    var="param" if purpose == "function" else "cvar", name=name
                ),
                "type {name}".format(name=name),
            )
        )

        return "\n".join(
            map(
                indent_all_but_first,
                map(
                    _fill,
                    filter(
                        None,
                        (
                            ":{key}: {doc}".format(
                                key=key,
                                doc=set_default_doc(
                                    (name, _param), emit_default_doc=emit_default_doc
                                )[1]["doc"].lstrip(),
                            )
                            if emit_doc and _param.get("doc")
                            else None,
                            ":{key_typ}: ```{typ}```".format(
                                key_typ=key_typ, typ=_param["typ"]
                            )
                            if emit_type and _param.get("typ")
                            else None,
                        ),
                    ),
                ),
            )
        )
    elif style == "numpydoc":
        return "\n".join(
            filter(
                None,
                (
                    _fill(
                        (_param["typ"] if _param.get("typ") else None)
                        if name == "return_type"
                        else "{name} :{typ}".format(
                            name=name,
                            typ=" {typ}".format(typ=_param["typ"])
                            if _param.get("typ")
                            else "",
                        )
                    )
                    if emit_type and _param.get("typ")
                    else None,
                    _fill(
                        indent(
                            set_default_doc(
                                (name, _param), emit_default_doc=emit_default_doc
                            )[1]["doc"],
                            tab,
                        )
                    )
                    if emit_doc and _param.get("doc")
                    else None,
                ),
            )
        )
    else:
        return "".join(
            filter(
                None,
                (
                    (
                        "  {typ}:".format(typ=_param["typ"])
                        if _param.get("typ")
                        else None
                    )
                    if name == "return_type"
                    else "  {name} ({typ}): ".format(
                        name=name,
                        typ="{typ!s}".format(typ=_param["typ"])
                        if _param.get("typ")
                        else "",
                    )
                    if _param.get("typ")
                    else "  {name}: ".format(name=name),
                    "{nl}{tab}{doc}".format(
                        doc=set_default_doc(
                            (name, _param), emit_default_doc=emit_default_doc
                        )[1]["doc"],
                        **{"nl": "\n", "tab": " " * 3}
                        if name == "return_type"
                        else {"nl": "", "tab": ""}
                    )
                    if emit_doc and _param.get("doc")
                    else None,
                ),
            )
        )


####################################################################
# internal functions for `parse_docstring_into_header_args_footer` #
####################################################################


def _get_token_start_idx(doc_str):
    """
    Get the start index of the token, minus starting whitespace for that line

    :param doc_str: The docstring
    :type doc_str: ```str```

    :return: index of token start or -1 if not found
    :rtype: ```int```
    """
    stack = []
    for idx, ch in enumerate(doc_str):
        if ch == "\n":
            indent_amount = count_iter_items(takewhile(str.isspace, stack))
            line = "".join(stack[indent_amount:])
            if line in NUMPYDOC_TOKENS_SET:
                i = indent_amount + idx + 1
                next_line = doc_str[i : doc_str.find("\n", i)]
                if next_line.count("-") == len(next_line):
                    return idx - len(stack)
            elif any(filter(line.startswith, TOKENS_SET)):
                return idx - len(stack)
            stack.clear()
        else:
            stack.append(ch)
    return -1


def _last_doc_str_token(doc_str):
    """
    Get start index of last docstring token (e.g., the index of ':' in ':rtype')

    :param doc_str: The docstring
    :type doc_str: ```str```

    :return: last_found index
    :rtype: ```Optional[int]```
    """
    last_found, penultimate_stack, stack = None, [], []
    for i, ch in enumerate(doc_str):
        if ch.isspace():
            if stack:
                if stack.count("-") == len(stack):
                    if "".join(penultimate_stack) in NUMPYDOC_TOKENS_SET:
                        last_found = i - len(stack) + len(penultimate_stack)
                elif "".join(stack) in TOKENS_SET:
                    last_found = i - len(stack)
                penultimate_stack = stack.copy()
                stack.clear()
        else:
            stack.append(ch)

    return last_found


def _get_start_of_last_found(last_found, doc_str):
    """
    Get last token found in doc_str…
    '  :params foo' returns index to first space ' '

    :param last_found: index of token start or None if not found
    :type last_found: ```Optional[int]```

    :param doc_str: The docstring
    :type doc_str: ```str```

    :return: _last_found index
    :rtype: ```Optional[int]```
    """
    last_found_starts = None

    for last_found_starts in range(last_found - 1, 0, -1):
        if doc_str[last_found_starts] == "\n":
            last_found_starts += 1
            break

    return last_found_starts


def _get_end_of_last_found_numpydoc(last_found, last_found_starts, doc_str):
    """
    Numpydoc specialisation of `_get_end_of_last_found` function

    :param last_found: index of token start or None if not found
    :type last_found: ```Optional[int]```

    :param last_found_starts: index of start of line containing token start
    :type last_found_starts: ```int```

    :param doc_str: The docstring
    :type doc_str: ```str```

    :return: _last_found index
    :rtype: ```Optional[int]```
    """
    countdown_from = last_found_starts - 1
    new_last_found_starts = None
    for new_last_found_starts in range(countdown_from - 1, 0, -1):
        if doc_str[new_last_found_starts] == "\n":
            new_last_found_starts += 1
            break

    if doc_str[new_last_found_starts:countdown_from] in NUMPYDOC_TOKENS_SET:
        # Find index of 'a' from last string in this format:
        # "arg_name : single/multi line desc"
        last_token_appearance = None
        stack = []
        for idx in range(last_found, len(doc_str)):
            if doc_str[idx] == "\n":
                for ch in stack:
                    if ch == ":":
                        last_token_appearance = idx
                        break
                stack.clear()
            else:
                stack.append(doc_str[idx])
        if last_token_appearance is not None:
            last_found_starts = last_token_appearance
            for i in range(last_token_appearance, 0, -1):
                if doc_str[i] == "\n":
                    last_found_starts = i + 1
                    break

            return last_found_starts


def _get_end_of_last_found(last_found, last_found_starts, doc_str, docstring_format):
    """
    '  :params foo' returns index to second 'o' + nl

    :param last_found: index of token start or None if not found
    :type last_found: ```Optional[int]```

    :param last_found_starts: index of start of line containing token start
    :type last_found_starts: ```int```

    :param doc_str: The docstring
    :type doc_str: ```str```

    :param docstring_format: Docstring format
    :type docstring_format: ```Style```

    :return: _last_found index
    :rtype: ```Optional[int]```
    """
    last_found_ends = None
    for last_found_ends in range(last_found, len(doc_str), 1):
        if doc_str[last_found_ends] == "\n":
            break
    last_found_ends += 1

    if docstring_format is Style.numpydoc and doc_str[
        last_found_starts:last_found
    ].count("-") == len(doc_str[last_found_starts:last_found]):
        return _get_end_of_last_found_numpydoc(
            last_found, 0 if last_found_starts is None else last_found_starts, doc_str
        )

    return last_found_ends


def _find_end_of_args_returns(last_found_ends, doc_str):
    """
    Handle multiline indented after keyword, e.g.,
      Returns:
         Foo
           more foo

    :param last_found_ends: index of end of line containing token start
    :type last_found_ends: ```int```

    :param doc_str: The docstring
    :type doc_str: ```str```

    :return: end of args_returns index
    :rtype: ```int```
    """
    smallest_indent = count_iter_items(
        takewhile(str.isspace, doc_str[last_found_ends:])
    )

    # Early exit… if current line isn't indented then it can't be a multiline arg/return descriptor
    if smallest_indent == 0 and last_found_ends is not None:
        return last_found_ends - 1

    return len(doc_str) - 1


def _get_token_last_idx_if_no_next_token(doc_str, last_found_starts):
    """
    Get the last index of the token (to end of line including newline)
    if no next token is detected (internal, only use is in `_get_token_last_idx`)

    :param doc_str: The docstring
    :type doc_str: ```str```

    :param last_found_starts: Index of where the last found token starts in `doc_str`
    :type last_found_starts: ```int```

    :return: index of token last or `None` if not found
    :rtype: ```int```
    """
    next_nl = last_found_starts + count_iter_items(
        takewhile(partial(ne, "\n"), doc_str[last_found_starts:])
    )

    next_line = doc_str[last_found_starts:next_nl]
    if frozenset(next_line) == frozenset(("-",)):
        line_start = line_end = next_nl + 1
        line_no = 0
        # last_space = line_no, line_start, line_end
        PrevParam = namedtuple(
            "PrevParam", ("line_no", "indent", "line_start", "line_end")
        )
        prev_param = PrevParam(
            line_no=None, indent=None, line_start=line_start, line_end=line_end
        )

        while line_end < len(doc_str):
            line_end += count_iter_items(
                takewhile(partial(ne, "\n"), doc_str[line_start:])
            )
            line = doc_str[line_start:line_end]
            line_no += 1

            if line.isspace():
                pass
            elif line_no - 2 == prev_param[0]:
                starting_whitespace = count_iter_items(takewhile(str.isspace, line))
                if starting_whitespace >= prev_param[1]:
                    # Handle multiline description of param
                    prev_param = PrevParam(
                        line_no=line_no,
                        indent=starting_whitespace,
                        line_start=line_start,
                        line_end=line_end,
                    )
            elif ":" in line:
                # This is a param
                prev_param = PrevParam(
                    line_no=line_no,
                    indent=count_iter_items(takewhile(str.isspace, line)),
                    line_start=line_start,
                    line_end=line_end,
                )

            line_start = line_end
            line_end += 1
        return prev_param.line_end + 1

        # if i_started_at == line_end:

    if doc_str[last_found_starts:next_nl] == "Raises:":
        # Don't treat this as anything special… short circuit
        return last_found_starts - 1


def _get_token_last_idx(doc_str):
    """
    Get the last index of the token (to end of line including newline)

    :param doc_str: The docstring
    :type doc_str: ```str```

    :return: index of token last or `len(doc_str) - 1` if not found
    :rtype: ```int```
    """

    last_found = _last_doc_str_token(doc_str)
    if last_found is None:
        return -1

    docstring_format = derive_docstring_format(doc_str)

    last_found_starts = _get_start_of_last_found(last_found, doc_str)
    last_found_ends = _get_end_of_last_found(
        last_found, last_found_starts, doc_str, docstring_format
    )

    idx = _find_end_of_args_returns(last_found_ends, doc_str)

    # Munch until previous nl
    while idx != 0 and doc_str[idx] != "\n":
        idx -= 1

    indent_amount = count_iter_items(takewhile(str.isspace, doc_str[idx + 1 :]))

    i_started_at = i = indent_amount + idx + 1
    if any(filter(doc_str[i:].startswith, TOKENS_SET)):
        # Munch until nl
        i += 1
        while i < len(doc_str) and doc_str[i] != "\n":
            i += 1

    if i_started_at == i:
        last_idx = _get_token_last_idx_if_no_next_token(
            doc_str, 0 if last_found_starts is None else last_found_starts
        )
        if last_idx is not None:
            return last_idx

    return i


########################################################################
# end internal functions for `parse_docstring_into_header_args_footer` #
########################################################################


def parse_docstring_into_header_args_footer(current_doc_str, original_doc_str):
    """
    Parse docstring into three parts: header; args|returns; footer

    :param current_doc_str: The current doc_str
    :type current_doc_str: ```str```

    :param original_doc_str: The original doc_str
    :type original_doc_str: ```str```

    :return: Header, args|returns, footer
    :rtype: ```Tuple[Optional[str], Optional[str], Optional[str]]```
    """
    # if not current_doc_str and not original_doc_str: return None, None, None

    # To quieten linter
    header_original = footer_original = None
    start_idx_current = last_idx_current = start_idx_original = last_idx_original = None

    if current_doc_str:
        start_idx_current = _get_token_start_idx(current_doc_str)
        last_idx_current = _get_token_last_idx(current_doc_str)
        # footer_current = (
        #     current_doc_str[last_idx_current:] if last_idx_current != -1 else None
        # )
        # header_current = (
        #     original_doc_str[:start_idx_current] if start_idx_current > -1 else None
        # )
        #
        # args_returns_current = current_doc_str[
        #     slice(
        #         start_idx_current if start_idx_current > -1 else None,
        #         last_idx_current if last_idx_current > -1 else None,
        #     )
        # ]

    if original_doc_str:
        start_idx_original = _get_token_start_idx(original_doc_str)
        last_idx_original = _get_token_last_idx(original_doc_str)
        footer_original = (
            original_doc_str[last_idx_original:] if last_idx_original != -1 else None
        )
        header_original = (
            original_doc_str[:start_idx_original] if start_idx_original > -1 else None
        )

        args_returns_original = original_doc_str[
            slice(
                start_idx_original if start_idx_original > -1 else None,
                last_idx_original if last_idx_original > -1 else None,
            )
        ]

    # Now we know where the args/returns were, and where they are now
    # To avoid whitespace issues, only copy across the args/returns portion, keep rest as original

    args_returns_current, args_returns_original = map(
        lambda doc_start_end: doc_start_end[0][
            slice(
                doc_start_end[1]
                if doc_start_end[1] is not None and doc_start_end[1] > -1
                else None,
                doc_start_end[2]
                if doc_start_end[2] is not None and doc_start_end[2] > -1
                else None,
            )
        ],
        (
            (current_doc_str, start_idx_current, last_idx_current),
            (original_doc_str, start_idx_original, last_idx_original),
        ),
    )

    indent_args_returns_original = count_iter_items(
        takewhile(str.isspace, args_returns_original or iter(()))
    )
    if indent_args_returns_original > 1 and args_returns_current:
        args_returns_current = indent(
            args_returns_current,
            prefix=" " * indent_args_returns_original,
            predicate=lambda _: _,
        )

    return header_original, args_returns_current, footer_original


def ensure_doc_args_whence_original(current_doc_str, original_doc_str):
    """
    Ensure doc args appear where they appeared originally

    :param current_doc_str: The current doc_str
    :type current_doc_str: ```str```

    :param original_doc_str: The original doc_str
    :type original_doc_str: ```str```

    :return: reshuffled doc_str with args/returns in same place as original (same header, footer, and whitespace)
    :rtype: ```str```
    """
    if original_doc_str and eq(
        *map(omit_whitespace, (original_doc_str, current_doc_str))
    ):
        return original_doc_str
    (
        original_header,
        current_args_returns,
        original_footer,
    ) = parse_docstring_into_header_args_footer(current_doc_str, original_doc_str)

    return header_args_footer_to_str(
        header=original_header or "",
        args_returns=current_args_returns or "",
        footer=original_footer or "",
    )


def header_args_footer_to_str(header, args_returns, footer):
    """
    Ensure there is always two newlines between header and next, and one between non-empty: args_returns and footer

    :param header: Header section
    :type header: ```str```

    :param args_returns: args|returns section
    :type args_returns: ```str```

    :param footer: Footer section
    :type footer: ```str```

    :return: One string with these three section combined with a minimum of one nl betwixt each and two at first
    :rtype: ```str```
    """
    header_end_nls = num_of_nls(header, end=True) if header else 0
    if args_returns:
        args_returns_start_nls = num_of_nls(args_returns, end=False)
        args_returns_ends_nls = num_of_nls(args_returns, end=True)
        args_returns = "{maybe_nls0}{args_returns}{maybe_nl1}".format(
            maybe_nls0="\n" * (args_returns_start_nls or 2)
            if args_returns_start_nls < 2 and header and not header_end_nls
            else "",
            args_returns=args_returns,
            maybe_nl1="\n" if not args_returns_ends_nls else "",
        )
        args_returns_start_nls = num_of_nls(args_returns, end=False)
        args_returns_ends_nls = num_of_nls(args_returns, end=True)
    else:
        args_returns_start_nls = args_returns_ends_nls = 0
    if footer:
        footer_start_nls = (
            count_chars_from(footer, str.isspace, "\n", end=False)
            or args_returns_ends_nls
        )
        # foot_end_has_nl = footer[-1] == "\n"
    else:
        footer_start_nls = 0  # foot_end_has_nl

    # Match indent of args_returns to header or footer
    if args_returns:
        header_or_footer = header if header else footer
        indent_amount = count_iter_items(takewhile(str.isspace, header_or_footer))
        newlines = (
            header_or_footer[:indent_amount].count("\n") if header_or_footer else 0
        )
        indent_amount = indent_amount - newlines
        current_indent_amount = count_iter_items(takewhile(str.isspace, args_returns))
        if current_indent_amount != indent_amount:
            _indent = indent_amount * " "
            len_args_returns = len(args_returns)
            args_returns = indent(args_returns, _indent, predicate=lambda _: _)
            if args_returns[-1] == "\n" and len_args_returns > 1:
                args_returns += _indent

    nls_after_header = header_end_nls + args_returns_start_nls
    nls_needed_after_header = (
        0
        if (nls_after_header > 1 or not header or not args_returns)
        else 1
        if nls_after_header == 1
        else 2
        if nls_after_header == 0
        else 0
    )

    return "{header}{maybe_nl0}{args_returns}{maybe_nl1}{footer}{maybe_nl2}".format(
        header=header,
        maybe_nl0="\n" * nls_needed_after_header,
        args_returns=args_returns,
        maybe_nl1=(
            "\n"
            if args_returns
            and footer
            and not footer_start_nls
            and not args_returns_ends_nls
            else ""
        ),
        footer=footer,
        maybe_nl2="",  # if foot_end_has_nl else "\n",
    )


class Style(Enum):
    """
    Simple enum taken from the docstring_parser codebase
    """

    rest = 1
    google = 2
    numpydoc = 3
    auto = 255


def derive_docstring_format(docstring):
    """
    Infer the docstring format of the provided docstring

    :param docstring: the docstring
    :type docstring: ```Optional[str]```

    :return: the style of docstring
    :rtype: ```Literal['rest', 'numpydoc', 'google']```
    """
    if docstring is None or any(map(partial(contains, docstring), TOKENS.rest)):
        style = Style.rest
    elif any(map(partial(contains, docstring), TOKENS.google)):
        style = Style.google
    else:
        style = Style.numpydoc
    return style


DOCSTRING_FORMATS = "rest", "google", "numpydoc"
Tokens = namedtuple("Tokens", DOCSTRING_FORMATS)
TOKENS = Tokens(
    (":param", ":cvar", ":ivar", ":var", ":type", ":raises", ":return", ":rtype"),
    ("Args:", "Kwargs:", "Raises:", "Returns:"),
    ("Parameters\n" "----------", "Returns\n" "-------"),
)
# Note: FPs possible for `"Parameters"` and `"Returns"` randomly thrown into normal doc_str
TOKENS_SET = frozenset(
    map(
        itemgetter(0),
        map(
            rpartial(str.partition, "\n"),
            chain.from_iterable(TOKENS._asdict().values()),
        ),
    )
)
NUMPYDOC_TOKENS_SET = frozenset(
    map(
        itemgetter(0),
        map(
            rpartial(str.partition, "\n"),
            TOKENS.numpydoc,
        ),
    )
)
ARG_TOKENS = Tokens(
    TOKENS.rest[:-2],
    (TOKENS.google[0],),
    (TOKENS.numpydoc[0],),
)
RETURN_TOKENS = Tokens(TOKENS.rest[-2:], (TOKENS.google[-1],), (TOKENS.numpydoc[-1],))


__all__ = [
    "ARG_TOKENS",
    "RETURN_TOKENS",
    "Style",
    "TOKENS",
    "derive_docstring_format",
    "emit_param_str",
    "ensure_doc_args_whence_original",
    "parse_docstring_into_header_args_footer",
]



"""
Functions which produce intermediate_repr from various different inputs
"""

import ast
from importlib import import_module


def get_internal_body(target_name, target_type, intermediate_repr):
    """
    Get the internal body from our IR

    :param target_name: name of target. If both `target_name` and `target_type` match internal body extract, then emit
    :type target_name: ```str```

    :param target_type: Type of target, static is static or global method, others just become first arg
    :type target_type: ```Literal['self', 'cls', 'static']```

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "_internal": {'body': List[ast.AST]},
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :return: Internal body or an empty tuple
    :rtype: ```Union[list, tuple]```
    """
    return (
        intermediate_repr["_internal"]["body"]
        if intermediate_repr.get("_internal", {}).get("body")
        and intermediate_repr["_internal"]["from_name"] == target_name
        and intermediate_repr["_internal"]["from_type"] == target_type
        else tuple()
    )


def ast_parse_fix(s):
    """
    Hack to resolve unbalanced parentheses SyntaxError acquired from PyTorch parsing
    TODO: remove

    :param s: String to parse
    :type s: ```str```

    :return: Value
    """
    balanced = (s.count("[") + s.count("]")) & 1 == 0
    return ast.parse(s if balanced else "{}]".format(s)).body[0].value


# def normalise_intermediate_representation(intermediate_repr):
#     """
#     Normalise the intermediate representation. Performs:
#     - Move header and footer of docstring to same place—and with same whitespace—as original docstring
#
#     :param intermediate_repr: a dictionary of form
#         {  "name": Optional[str],
#            "type": Optional[str],
#            "doc": Optional[str],
#            "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
#            "returns": Optional[OrderedDict[Literal['return_type'],
#                                            {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
#     :type intermediate_repr: ```dict```
#
#     :return: a dictionary of form
#         {  "name": Optional[str],
#            "type": Optional[str],
#            "doc": Optional[str],
#            "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
#            "returns": Optional[OrderedDict[Literal['return_type'],
#                                            {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
#     :rtype: ```dict```
#     """
#     current_doc_str = intermediate_repr["doc"]
#     original_doc_str = intermediate_repr.get("_internal", {"original_doc_str": None})[
#         "original_doc_str"
#     ]
#     intermediate_repr["doc"] = ensure_doc_args_whence_original(
#         current_doc_str=current_doc_str, original_doc_str=original_doc_str
#     )
#     return intermediate_repr


def get_emitter(emit_name):
    """
    Get emiter function specialised for output `node`

    :param emit_name: Which type to emit.
    :type emit_name: ```Literal["argparse", "class", "function", "json_schema",
                                 "pydantic", "sqlalchemy", "sqlalchemy_table"]```

    :return: Function which returns intermediate_repr
    :rtype: ```Callable[[...], dict]````
    """
    emit_name = {"class": "class_"}.get(emit_name, emit_name)
    return getattr(
        import_module(
            ".".join(
                (
                    "cdd",
                    "sqlalchemy" if emit_name == "sqlalchemy_table" else emit_name,
                    "emit",
                )
            )
        ),
        emit_name,
    )


__all__ = [
    "ast_parse_fix",
    "get_internal_body",
    "get_emitter"
    # "normalise_intermediate_representation",
]



"""
Pydantic `class` emitter

https://pydantic-docs.helpmanual.io/usage/schema/
"""

from functools import partial

import cdd.class_.emit

pydantic = partial(cdd.class_.emit.class_, class_bases=("BaseModel",))

__all__ = ["pydantic"]



"""
Tests for `cdd.emit.function`
"""

import ast
from ast import FunctionDef
from collections import OrderedDict
from copy import deepcopy
from functools import partial
from textwrap import indent
from unittest import TestCase

import cdd.argparse_function.emit
import cdd.argparse_function.parse
import cdd.class_.emit
import cdd.class_.parse
import cdd.docstring.emit
import cdd.docstring.parse
import cdd.function.emit
import cdd.function.parse
import cdd.json_schema.emit
import cdd.shared.emit.file
import cdd.sqlalchemy.emit
from cdd.shared.ast_utils import get_function_type, set_value
from cdd.shared.pure_utils import none_types, rpartial, tab
from cdd.tests.mocks.classes import class_squared_hinge_config_ast
from cdd.tests.mocks.docstrings import (
    docstring_header_str,
    docstring_no_type_no_default_tpl_str,
    docstring_str,
)
from cdd.tests.mocks.methods import (
    class_with_method_ast,
    class_with_method_str,
    class_with_method_types_ast,
    class_with_method_types_str,
    function_google_tf_squared_hinge_str,
)
from cdd.tests.utils_for_tests import reindent_docstring, run_ast_test, unittest_main


class TestEmitFunction(TestCase):
    """Tests emission"""

    def test_to_function(self) -> None:
        """
        Tests whether `function` produces method from `class_with_method_types_ast` given `docstring_str`
        """

        function_def = reindent_docstring(
            deepcopy(
                next(
                    filter(
                        rpartial(isinstance, FunctionDef),
                        class_with_method_types_ast.body,
                    )
                )
            )
        )

        function_name = function_def.name
        function_type = get_function_type(function_def)

        gen_ast = cdd.function.emit.function(
            cdd.docstring.parse.docstring(docstring_str),
            function_name=function_name,
            function_type=function_type,
            emit_default_doc=False,
            type_annotations=True,
            emit_separating_tab=True,
            indent_level=1,
            emit_as_kwonlyargs=False,
        )

        run_ast_test(
            self,
            gen_ast=gen_ast,
            gold=function_def,
        )

    def test_to_function_with_docstring_types(self) -> None:
        """
        Tests that `function` can generate a function_def with types in docstring
        """

        # Sanity check
        run_ast_test(
            self,
            class_with_method_ast,
            gold=ast.parse(class_with_method_str).body[0],
        )

        function_def = reindent_docstring(
            deepcopy(
                next(
                    filter(
                        rpartial(isinstance, FunctionDef), class_with_method_ast.body
                    )
                )
            )
        )

        ir = cdd.function.parse.function(function_def)
        gen_ast = reindent_docstring(
            cdd.function.emit.function(
                ir,
                function_name=function_def.name,
                function_type=get_function_type(function_def),
                emit_default_doc=False,
                type_annotations=False,
                indent_level=1,
                emit_separating_tab=True,
                emit_as_kwonlyargs=False,
                word_wrap=False,
            )
        )

        run_ast_test(self, gen_ast=gen_ast, gold=function_def)

    def test_to_function_with_type_annotations(self) -> None:
        """
        Tests that `function` can generate a function_def with inline types
        """
        function_def = deepcopy(
            next(
                filter(
                    rpartial(isinstance, FunctionDef), class_with_method_types_ast.body
                )
            )
        )
        function_name = function_def.name
        function_type = get_function_type(function_def)
        function_def.body[0].value = set_value(
            "\n{tab}{ds}{tab}".format(
                tab=tab,
                ds=indent(
                    docstring_no_type_no_default_tpl_str.format(
                        header_doc_str=docstring_header_str
                    ),
                    prefix=tab,
                    predicate=lambda _: _,
                ).lstrip(),
            )
        )

        gen_ast = cdd.function.emit.function(
            cdd.function.parse.function(
                function_def,
                function_name=function_name,
                function_type=function_type,
            ),
            function_name=function_name,
            function_type=function_type,
            emit_default_doc=False,
            type_annotations=True,
            emit_separating_tab=True,
            indent_level=0,
            emit_as_kwonlyargs=False,
        )
        # emit.file(gen_ast, os.path.join(os.path.dirname(__file__), 'delme.py'), mode='wt')

        run_ast_test(
            self,
            gen_ast=gen_ast,
            gold=function_def,
        )

    def test_to_function_emit_as_kwonlyargs(self) -> None:
        """
        Tests whether `function` produces method with keyword only arguments
        """
        function_def = reindent_docstring(
            deepcopy(
                next(
                    filter(
                        rpartial(isinstance, FunctionDef),
                        ast.parse(
                            class_with_method_types_str.replace("self", "self, *")
                        )
                        .body[0]
                        .body,
                    )
                )
            )
        )
        function_name = function_def.name
        function_type = get_function_type(function_def)

        gen_ast = cdd.function.emit.function(
            cdd.docstring.parse.docstring(docstring_str),
            function_name=function_name,
            function_type=function_type,
            emit_default_doc=False,
            type_annotations=True,
            emit_separating_tab=True,
            indent_level=1,
            emit_as_kwonlyargs=True,
        )

        run_ast_test(
            self,
            gen_ast=gen_ast,
            gold=function_def,
        )

    def test_from_function_google_tf_squared_hinge_str_to_class(self) -> None:
        """
        Tests that `emit.function` produces correctly with:
        - __call__
        """

        gold_ir = cdd.class_.parse.class_(class_squared_hinge_config_ast)
        gold_ir.update(
            {
                key: OrderedDict(
                    (
                        (
                            name,
                            {
                                k: v
                                for k, v in _param.items()
                                if k != "typ"
                                and (k != "default" or v not in none_types)
                            },
                        )
                        for name, _param in gold_ir[key].items()
                    )
                )
                for key in ("params", "returns")
            }
        )

        gen_ir = cdd.function.parse.function(
            ast.parse(function_google_tf_squared_hinge_str).body[0],
            infer_type=True,
            word_wrap=False,
        )
        self.assertEqual(
            *map(lambda ir: ir["returns"]["return_type"]["doc"], (gen_ir, gold_ir))
        )
        # print('#gen_ir')
        # pp(gen_ir)
        # print('#gold_ir')
        # pp(gold_ir)

        run_ast_test(
            self,
            *map(
                partial(
                    cdd.class_.emit.class_,
                    class_name="SquaredHingeConfig",
                    emit_call=True,
                    emit_default_doc=True,
                    emit_original_whitespace=True,
                    word_wrap=False,
                ),
                (gen_ir, gold_ir),
            )
        )


unittest_main()



"""
Mocks for the argparse function
"""

from ast import (
    Assign,
    Attribute,
    BinOp,
    Call,
    Expr,
    FunctionDef,
    If,
    Load,
    Mult,
    Name,
    Return,
    Store,
    Tuple,
    arguments,
    fix_missing_locations,
    keyword,
)

from cdd.shared.ast_utils import FALLBACK_TYP, maybe_type_comment, set_arg, set_value
from cdd.shared.pure_utils import deindent, emit_separating_tabs, tab
from cdd.tests.mocks.classes import (
    class_torch_nn_l1loss_docstring_str,
    tensorboard_doc_str_no_args_str,
)
from cdd.tests.mocks.docstrings import docstring_header_no_nl_str, docstring_header_str
from cdd.tests.utils_for_tests import remove_args_from_docstring

argparse_add_argument_ast = Expr(
    Call(
        args=[set_value("--num")],
        func=Attribute(
            Name("argument_parser", Load()),
            "add_argument",
            Load(),
        ),
        keywords=[
            keyword(arg="type", value=Name("int", Load()), identifier=None),
            keyword(
                arg="required",
                value=set_value(True),
                identifier=None,
            ),
        ],
        expr=None,
        expr_func=None,
    )
)

_argparse_doc_str_tuple = (
    "Set CLI arguments\n",
    ":param argument_parser: argument parser",
    ":type argument_parser: ```ArgumentParser```\n",
)

_cli_doc_str = "\n{tab}".format(tab=tab).join(
    _argparse_doc_str_tuple
    + (
        ":return: argument_parser, Train and tests dataset splits.",
        ":rtype: ```Tuple[ArgumentParser, Union[Tuple[tf.data.Dataset, tf.data.Dataset],"
        " Tuple[np.ndarray,\n{tab}np.ndarray]]]```".format(tab=tab),
    )
)

_argparse_doc_tuple = (
    _argparse_doc_str_tuple[0].rstrip("\n"),
    "",
    _argparse_doc_str_tuple[1],
    _argparse_doc_str_tuple[2],
    ":return: argument_parser, Train and tests dataset splits.",
    ":rtype: ```Tuple[ArgumentParser, Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray,",
    "np.ndarray]]]```",
)

_argparse_doc_str = "\n".join(_argparse_doc_tuple)

_argparse_description_str = emit_separating_tabs(
    "\n{tab}".format(tab=tab).join(
        _argparse_doc_tuple[:-1]
        + (
            "{tab}{last_line_of_doc_str}".format(
                tab=tab, last_line_of_doc_str=_argparse_doc_tuple[-1]
            ),
        )
    )
)
_argparse_doc_stripped_str = _argparse_doc_str.replace(
    "{tab}\n{tab}".format(tab=tab), ""
)

_cli_doc_expr = Expr(set_value(_argparse_description_str))

_cli_doc_nosplit_str = emit_separating_tabs(
    "\n{tab}".format(tab=tab).join(
        _argparse_doc_str_tuple
        + (
            ":return: argument_parser",
            ":rtype: ```ArgumentParser```",
        )
    )
).strip(" \n")

_cli_doc_nosplit_expr = Expr(
    set_value(
        "\n{tab}{_cli_doc_nosplit_str}\n{tab}".format(
            _cli_doc_nosplit_str=_cli_doc_nosplit_str, tab=tab
        )
    )
)

as_numpy_argparse_call = Expr(
    Call(
        args=[set_value("--as_numpy")],
        func=Attribute(
            Name("argument_parser", Load()),
            "add_argument",
            Load(),
        ),
        keywords=[
            keyword(arg="type", value=Name("bool", Load()), identifier=None),
            keyword(
                arg="help",
                value=set_value(
                    "Convert to numpy ndarrays",
                ),
                identifier=None,
            ),
        ],
        expr=None,
        expr_func=None,
    )
)

_argparse_add_arguments = (
    Expr(
        Call(
            args=[set_value("--dataset_name")],
            func=Attribute(
                Name("argument_parser", Load()),
                "add_argument",
                Load(),
            ),
            keywords=[
                keyword(
                    arg="help",
                    value=set_value("name of dataset."),
                    identifier=None,
                ),
                keyword(
                    arg="required",
                    value=set_value(True),
                    identifier=None,
                ),
                keyword(
                    arg="default",
                    value=set_value("mnist"),
                    identifier=None,
                ),
            ],
            expr=None,
            expr_func=None,
        )
    ),
    Expr(
        Call(
            args=[set_value("--tfds_dir")],
            func=Attribute(
                Name("argument_parser", Load()),
                "add_argument",
                Load(),
            ),
            keywords=[
                keyword(
                    arg="help",
                    value=set_value("directory to look for models in."),
                    identifier=None,
                ),
                keyword(
                    arg="required",
                    value=set_value(True),
                    identifier=None,
                ),
                keyword(
                    arg="default",
                    value=set_value("~/tensorflow_datasets"),
                    identifier=None,
                ),
            ],
            expr=None,
            expr_func=None,
        )
    ),
    Expr(
        Call(
            args=[set_value("--K")],
            func=Attribute(
                Name("argument_parser", Load()),
                "add_argument",
                Load(),
            ),
            keywords=[
                keyword(
                    arg="choices",
                    value=Tuple(
                        ctx=Load(),
                        elts=list(map(set_value, ("np", "tf"))),
                        expr=None,
                    ),
                    identifier=None,
                ),
                keyword(
                    arg="help",
                    value=set_value(
                        "backend engine, e.g., `np` or `tf`.",
                    ),
                    identifier=None,
                ),
                keyword(
                    arg="required",
                    value=set_value(True),
                    identifier=None,
                ),
                keyword(
                    arg="default",
                    value=set_value("np"),
                    identifier=None,
                ),
            ],
            expr=None,
            expr_func=None,
        )
    ),
    as_numpy_argparse_call,
    Expr(
        Call(
            args=[set_value("--data_loader_kwargs")],
            func=Attribute(
                Name("argument_parser", Load()),
                "add_argument",
                Load(),
            ),
            keywords=[
                keyword(arg="type", value=Name("loads", Load()), identifier=None),
                keyword(
                    arg="help",
                    value=set_value(
                        "pass this as arguments to data_loader function",
                    ),
                    identifier=None,
                ),
            ],
            expr=None,
            expr_func=None,
        )
    ),
)

_argparse_return = Return(
    value=Tuple(
        ctx=Load(),
        elts=[
            Name("argument_parser", Load()),
            Tuple(
                ctx=Load(),
                elts=[
                    Call(
                        args=[set_value(0)],
                        func=Attribute(
                            Name("np", Load()),
                            "empty",
                            Load(),
                        ),
                        keywords=[],
                        expr=None,
                        expr_func=None,
                    )
                ]
                * 2,
                expr=None,
            ),
        ],
        expr=None,
    ),
    expr=None,
)

argparse_func_str = '''
def set_cli_args(argument_parser):
    """
    {_cli_doc_str}
    """
    argument_parser.description = (
        {description!r}
    )
    argument_parser.add_argument(
        "--dataset_name",
        help="name of dataset.",
        required=True,
        default="mnist",
    )
    argument_parser.add_argument(
        "--tfds_dir",
        help="directory to look for models in.",
        required=True,
        default="~/tensorflow_datasets",
    )
    argument_parser.add_argument(
        "--K",
        choices=("np", "tf"),
        help="backend engine, e.g., `np` or `tf`.",
        required=True,
        default="np",
    )
    argument_parser.add_argument(
        "--as_numpy", type=bool, help="Convert to numpy ndarrays",
    )
    argument_parser.add_argument(
        "--data_loader_kwargs",
        type=loads,
        help="pass this as arguments to data_loader function",
    )
    return argument_parser, (np.empty(0), np.empty(0))
'''.format(
    _cli_doc_str=_cli_doc_str,
    description=docstring_header_str.strip(),
)

argparse_func_with_body_str = '''
def set_cli_args(argument_parser):
    """
    {_cli_doc_str}
    """
    argument_parser.description = (
        {header_doc_str!r}
    )
    argument_parser.add_argument(
        '--dataset_name', type=str, help='name of dataset.', required=True, default='mnist'
    )
    argument_parser.add_argument(
        '--tfds_dir',
        type=str,
        help='directory to look for models in.',
        default='~/tensorflow_datasets',
    )
    argument_parser.add_argument(
        '--K',
        type={FALLBACK_TYP},
        choices=('np', 'tf'),
        help='backend engine, e.g., `np` or `tf`.',
        required=True,
        default='np',
    )
    argument_parser.add_argument('--as_numpy', type=bool, help='Convert to numpy ndarrays')
    argument_parser.add_argument(
        '--data_loader_kwargs', type=loads, help='pass this as arguments to data_loader function'
    )
    # some comment
    print(5*5)
    if True:
        print(True)
        return 5
    return argument_parser, (np.empty(0), np.empty(0))
'''.format(
    _cli_doc_str=_cli_doc_str,
    FALLBACK_TYP=FALLBACK_TYP,
    header_doc_str=docstring_header_str,
)

argparse_func_action_append_str = '''
def set_cli_action_append(argument_parser):
    """
    {_cli_doc_str}
    """
    argument_parser.description = {header_doc_str!r}
    argument_parser.add_argument(
        "--callbacks",
        type=str,
        choices=(
            "BaseLogger",
            "CSVLogger",
            "Callback",
            "CallbackList",
            "EarlyStopping",
            "History",
            "LambdaCallback",
            "LearningRateScheduler",
            "ModelCheckpoint",
            "ProgbarLogger",
            "ReduceLROnPlateau",
            "RemoteMonitor",
            "TensorBoard",
            "TerminateOnNaN",
        ),
        action="append",
        help="Collection of callables that are run inside the training loop",
    )
    return argument_parser
'''.format(
    _cli_doc_str=_cli_doc_str, header_doc_str=docstring_header_str
)

argparse_func_ast = fix_missing_locations(
    FunctionDef(
        args=arguments(
            args=[set_arg("argument_parser")],
            defaults=[],
            kw_defaults=[],
            kwarg=None,
            kwonlyargs=[],
            posonlyargs=[],
            vararg=None,
            arg=None,
        ),
        body=[
            _cli_doc_expr,
            Assign(
                targets=[
                    Attribute(
                        Name("argument_parser", Load()),
                        "description",
                        Store(),
                    )
                ],
                value=set_value(docstring_header_no_nl_str),
                expr=None,
                **maybe_type_comment
            ),
            *_argparse_add_arguments,
            _argparse_return,
        ],
        decorator_list=[],
        name="set_cli_args",
        returns=None,
        arguments_args=None,
        stmt=None,
        identifier_name=None,
        **maybe_type_comment
    )
)

argparse_func_with_body_ast = fix_missing_locations(
    FunctionDef(
        args=arguments(
            args=[set_arg("argument_parser")],
            defaults=[],
            kw_defaults=[],
            kwarg=None,
            kwonlyargs=[],
            posonlyargs=[],
            vararg=None,
            arg=None,
        ),
        body=[
            _cli_doc_expr,
            Assign(
                targets=[
                    Attribute(
                        Name("argument_parser", Load()),
                        "description",
                        Store(),
                    )
                ],
                value=set_value(_argparse_description_str),
                expr=None,
                **maybe_type_comment
            ),
            *_argparse_add_arguments,
            Expr(
                Call(
                    args=[
                        BinOp(
                            set_value(5),
                            Mult(),
                            set_value(5),
                        )
                    ],
                    func=Name("print", Load()),
                    keywords=[],
                    expr=None,
                    expr_func=None,
                )
            ),
            If(
                body=[
                    Expr(
                        Call(
                            args=[set_value(True)],
                            func=Name("print", Load()),
                            keywords=[],
                            expr=None,
                            expr_func=None,
                        )
                    ),
                    Return(
                        value=set_value(5),
                        expr=None,
                    ),
                ],
                orelse=[],
                test=set_value(True),
                expr_test=None,
                stmt=None,
            ),
            _argparse_return,
        ],
        decorator_list=[],
        name="set_cli_args",
        returns=None,
        arguments_args=None,
        identifier_name=None,
        stmt=None,
        **maybe_type_comment
    )
)

argparse_func_action_append_ast = fix_missing_locations(
    FunctionDef(
        name="set_cli_action_append",
        args=arguments(
            posonlyargs=[],
            args=[set_arg("argument_parser")],
            kwonlyargs=[],
            kw_defaults=[],
            defaults=[],
            arg=None,
            vararg=None,
            kwarg=None,
        ),
        body=[
            _cli_doc_nosplit_expr,
            Assign(
                targets=[
                    Attribute(Name("argument_parser", Load()), "description", Store())
                ],
                value=set_value(docstring_header_no_nl_str),
                expr=None,
                **maybe_type_comment
            ),
            Expr(
                Call(
                    func=Attribute(
                        Name("argument_parser", Load()), "add_argument", Load()
                    ),
                    args=[set_value("--callbacks")],
                    keywords=[
                        keyword(
                            arg="type",
                            value=Name("str", Load()),
                            identifier=None,
                        ),
                        keyword(
                            arg="choices",
                            value=Tuple(
                                elts=list(
                                    map(
                                        set_value,
                                        (
                                            "BaseLogger",
                                            "CSVLogger",
                                            "Callback",
                                            "CallbackList",
                                            "EarlyStopping",
                                            "History",
                                            "LambdaCallback",
                                            "LearningRateScheduler",
                                            "ModelCheckpoint",
                                            "ProgbarLogger",
                                            "ReduceLROnPlateau",
                                            "RemoteMonitor",
                                            "TensorBoard",
                                            "TerminateOnNaN",
                                        ),
                                    )
                                ),
                                ctx=Load(),
                                expr=None,
                            ),
                            identifier=None,
                        ),
                        keyword(
                            arg="action",
                            value=set_value("append"),
                            identifier=None,
                        ),
                        keyword(
                            arg="help",
                            value=set_value(
                                "Collection of callables that are run inside the training loop"
                            ),
                            identifier=None,
                        ),
                    ],
                    expr=None,
                    expr_func=None,
                )
            ),
            Return(value=Name("argument_parser", Load()), expr=None),
        ],
        decorator_list=[],
        returns=None,
        arguments_args=None,
        identifier_name=None,
        stmt=None,
        **maybe_type_comment
    )
)

argparse_function_google_tf_tensorboard_ast = FunctionDef(
    args=arguments(
        args=[set_arg("argument_parser")],
        defaults=[],
        kw_defaults=[],
        kwarg=None,
        kwonlyargs=[],
        posonlyargs=[],
        vararg=None,
        arg=None,
    ),
    body=[
        _cli_doc_nosplit_expr,
        Assign(
            targets=[
                Attribute(
                    Name("argument_parser", Load()),
                    "description",
                    Store(),
                )
            ],
            lineno=None,
            value=set_value(tensorboard_doc_str_no_args_str),
            expr=None,
            **maybe_type_comment
        ),
        Expr(
            Call(
                args=[set_value("--log_dir")],
                func=Attribute(
                    Name("argument_parser", Load()),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(
                        arg="help",
                        value=set_value(
                            "the path of the directory where to save the log files to be parsed by TensorBoard."
                        ),
                        identifier=None,
                    ),
                    keyword(arg="required", value=set_value(True), identifier=None),
                    keyword(arg="default", value=set_value("logs"), identifier=None),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Expr(
            Call(
                args=[set_value("--histogram_freq")],
                func=Attribute(
                    Name("argument_parser", Load()),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(arg="type", value=Name("int", Load()), identifier=None),
                    keyword(
                        arg="help",
                        value=set_value(
                            "frequency (in epochs) at which to compute activation and weight histograms for the layers"
                            " of the model. If set to 0, histograms won't be computed. Validation data (or split) must"
                            " be specified for histogram visualizations."
                        ),
                        identifier=None,
                    ),
                    keyword(arg="required", value=set_value(True), identifier=None),
                    keyword(arg="default", value=set_value(0), identifier=None),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Expr(
            Call(
                args=[set_value("--write_graph")],
                func=Attribute(
                    Name("argument_parser", Load()),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(arg="type", value=Name("bool", Load()), identifier=None),
                    keyword(
                        arg="help",
                        value=set_value(
                            "whether to visualize the graph in TensorBoard. The log file can become quite large when"
                            " write_graph is set to True."
                        ),
                        identifier=None,
                    ),
                    keyword(arg="required", value=set_value(True), identifier=None),
                    keyword(arg="default", value=set_value(True), identifier=None),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Expr(
            Call(
                args=[set_value("--write_images")],
                func=Attribute(
                    Name("argument_parser", Load()),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(arg="type", value=Name("bool", Load()), identifier=None),
                    keyword(
                        arg="help",
                        value=set_value(
                            "whether to write model weights to visualize as image in TensorBoard."
                        ),
                        identifier=None,
                    ),
                    keyword(arg="required", value=set_value(True), identifier=None),
                    keyword(arg="default", value=set_value(False), identifier=None),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Expr(
            Call(
                args=[set_value("--update_freq")],
                func=Attribute(
                    Name("argument_parser", Load()),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(
                        arg="help",
                        value=set_value(
                            "`'batch'` or `'epoch'` or integer. When using `'batch'`, writes the losses and metrics "
                            "to TensorBoard after each batch. The same applies for `'epoch'`. If using an integer, "
                            "let's say `1000`, the callback will write the metrics and losses to TensorBoard every "
                            "1000 batches. Note that writing too frequently to TensorBoard can slow down your training"
                            "."
                        ),
                        identifier=None,
                    ),
                    keyword(arg="required", value=set_value(True), identifier=None),
                    keyword(arg="default", value=set_value("epoch"), identifier=None),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Expr(
            Call(
                args=[set_value("--profile_batch")],
                func=Attribute(
                    Name("argument_parser", Load()),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(arg="type", value=Name("int", Load()), identifier=None),
                    keyword(
                        arg="help",
                        value=set_value(
                            "Profile the batch(es) to sample compute characteristics. profile_batch must be a"
                            " non-negative integer or a tuple of integers. A pair of positive integers signify a range"
                            " of batches to profile. By default, it will profile the second batch. Set profile_batch=0"
                            " to disable profiling."
                        ),
                        identifier=None,
                    ),
                    keyword(arg="required", value=set_value(True), identifier=None),
                    keyword(arg="default", value=set_value(2), identifier=None),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Expr(
            Call(
                args=[set_value("--embeddings_freq")],
                func=Attribute(
                    Name("argument_parser", Load()),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(arg="type", value=Name("int", Load()), identifier=None),
                    keyword(
                        arg="help",
                        value=set_value(
                            "frequency (in epochs) at which embedding layers will be visualized. "
                            "If set to 0, embeddings won't be visualized."
                        ),
                        identifier=None,
                    ),
                    keyword(arg="required", value=set_value(True), identifier=None),
                    keyword(arg="default", value=set_value(0), identifier=None),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Expr(
            Call(
                args=[set_value("--embeddings_metadata")],
                func=Attribute(
                    Name("argument_parser", Load()),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(
                        arg="help",
                        value=set_value(
                            "a dictionary which maps layer name to a file name in which metadata for this"
                            " embedding layer is saved. See the "
                            "[details]( https://www.tensorflow.org/how_tos/embedding_viz/#metadata_optional) about"
                            " metadata files format. In case if the same metadata file is used for all embedding"
                            " layers, string can be passed."
                        ),
                        identifier=None,
                    ),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Return(value=Name("argument_parser", Load()), expr=None),
    ],
    decorator_list=[],
    name="set_cli_args",
    returns=None,
    arguments_args=None,
    identifier_name=None,
    stmt=None,
    lineno=None,
    **maybe_type_comment
)

argparse_func_torch_nn_l1loss_ast = FunctionDef(
    args=arguments(
        args=[set_arg("argument_parser")],
        defaults=[],
        kw_defaults=[],
        kwarg=None,
        kwonlyargs=[],
        posonlyargs=[],
        vararg=None,
        arg=None,
    ),
    body=[
        _cli_doc_nosplit_expr,
        Assign(
            targets=[
                Attribute(
                    Name("argument_parser", Load()),
                    "description",
                    Store(),
                )
            ],
            lineno=None,
            value=set_value(
                deindent(
                    remove_args_from_docstring(class_torch_nn_l1loss_docstring_str), 1
                )
            ),
            expr=None,
            **maybe_type_comment
        ),
        Expr(
            Call(
                args=[set_value("--size_average")],
                func=Attribute(
                    Name("argument_parser", Load()),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(
                        arg="type",
                        value=Name(
                            "bool",
                            Load(),
                        ),
                        identifier=None,
                    ),
                    keyword(
                        arg="help",
                        value=set_value(
                            "Deprecated (see :attr:`reduction`)."
                            " By default, the losses are averaged over each loss element in the batch. Note that for"
                            " some losses, there are multiple elements per sample. If the field :attr:`size_average`"
                            " is set to ``False``, the losses are instead summed for each minibatch. Ignored when"
                            " reduce is ``False``."
                        ),
                        identifier=None,
                    ),
                    keyword(arg="default", value=set_value(True), identifier=None),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Expr(
            Call(
                args=[set_value("--reduce")],
                func=Attribute(
                    Name("argument_parser", Load()),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(arg="type", value=Name("bool", Load()), identifier=None),
                    keyword(
                        arg="help",
                        value=set_value(
                            "Deprecated (see :attr:`reduction`). "
                            "By default, the losses are averaged or summed over observations for each minibatch "
                            "depending on :attr:`size_average`. When :attr:`reduce` is ``False``, returns a loss "
                            "per batch element instead and ignores :attr:`size_average`."
                        ),
                        identifier=None,
                    ),
                    keyword(arg="default", value=set_value(True), identifier=None),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Expr(
            Call(
                args=[set_value("--reduction")],
                func=Attribute(
                    Name("argument_parser", Load()),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(
                        arg="help",
                        value=set_value(
                            "Specifies the reduction to apply to the output:"
                            " ``'none'`` | ``'mean'`` | ``'sum'``. "
                            "``'none'``: no reduction will be applied, `"
                            "`'mean'``: the sum of the output will be divided by the number of elements in the output,"
                            " "
                            "``'sum'``: the output will be summed. "
                            "Note: :attr:`size_average` and :attr:`reduce` are in the process of being deprecated,"
                            " and in the meantime, specifying either of those two args will"
                            " override :attr:`reduction`."
                        ),
                        identifier=None,
                    ),
                    keyword(arg="default", value=set_value("mean"), identifier=None),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Expr(
            Call(
                args=[set_value("--__constants__")],
                func=Attribute(
                    Name(
                        "argument_parser",
                        Load(),
                    ),
                    "add_argument",
                    Load(),
                ),
                keywords=[
                    keyword(
                        arg="type",
                        value=Name(
                            "str",
                            Load(),
                        ),
                        identifier=None,
                    ),
                    keyword(arg="action", value=set_value("append"), identifier=None),
                    keyword(arg="required", value=set_value(True), identifier=None),
                    keyword(
                        arg="default", value=set_value("reduction"), identifier=None
                    ),
                ],
                expr=None,
                expr_func=None,
            )
        ),
        Return(
            value=Name(
                "argument_parser",
                Load(),
            ),
            expr=None,
        ),
    ],
    decorator_list=[],
    name="set_cli_args",
    returns=None,
    arguments_args=None,
    identifier_name=None,
    stmt=None,
    lineno=None,
    **maybe_type_comment
)

argparse_add_argument_expr = Expr(
    Call(
        args=[set_value("--byo")],
        func=Attribute(
            Name("argument_parser", Load()),
            "add_argument",
            Load(),
        ),
        keywords=[
            keyword(arg="type", value=Name("str", Load()), identifier=None),
            keyword(arg="action", value=set_value("append"), identifier=None),
            keyword(arg="required", value=set_value(True), identifier=None),
        ],
        expr=None,
        expr_func=None,
    )
)

__all__ = [
    "argparse_add_argument_ast",
    "argparse_add_argument_expr",
    "_argparse_doc_str_tuple",
    "argparse_func_action_append_ast",
    "argparse_func_action_append_str",
    "argparse_func_ast",
    "argparse_func_str",
    "argparse_func_torch_nn_l1loss_ast",
    "argparse_func_with_body_ast",
    "argparse_func_with_body_str",
    "argparse_function_google_tf_tensorboard_ast",
]



"""
Transform from string or AST representations of input, to intermediate_repr, a dictionary of form:
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
"""

from ast import AnnAssign, Assign, AsyncFunctionDef, ClassDef, FunctionDef

PARSERS = [
    "argparse_function",
    "class_",
    "docstring",
    "function",
    "json_schema",
    # "openapi",
    "pydantic",
    "routes",
    "sqlalchemy",
    "sqlalchemy_table",
]

kind2instance_type = {
    "argparse": (FunctionDef,),
    "argparse_function": (FunctionDef,),
    "class": (ClassDef,),
    "class_": (ClassDef,),
    "function": (FunctionDef, AsyncFunctionDef),
    "method": (FunctionDef, AsyncFunctionDef),
    "pydantic": (ClassDef,),
    "sqlalchemy_table": (Assign, AnnAssign),
    "sqlalchemy": (ClassDef,),
}

__all__ = ["PARSERS", "kind2instance_type"]



""" Tests for ast_cst_utils """

from ast import ClassDef, Expr, FunctionDef, Load, Name, Pass, walk
from copy import deepcopy
from io import StringIO
from operator import attrgetter
from os import path
from os.path import extsep
from unittest import TestCase
from unittest.mock import MagicMock, patch

from cdd.shared.ast_cst_utils import (
    Delta,
    debug_doctrans,
    find_cst_at_ast,
    maybe_replace_doc_str_in_function_or_class,
    maybe_replace_function_args,
    maybe_replace_function_return_type,
)
from cdd.shared.ast_utils import set_value
from cdd.shared.cst_utils import (
    ClassDefinitionStart,
    FunctionDefinitionStart,
    TripleQuoted,
    reindent_block_with_pass_body,
)
from cdd.shared.pure_utils import identity, rpartial
from cdd.shared.source_transformer import ast_parse
from cdd.tests.mocks.cst import cstify_cst
from cdd.tests.utils_for_tests import unittest_main


class TestAstCstUtils(TestCase):
    """Test class for cst_utils"""

    def setUp(self) -> None:
        """
        Initialise vars useful for multiple tests
        """
        with open(
            path.join(
                path.dirname(__file__),
                "../mocks",
                "cstify{extsep}py".format(extsep=extsep),
            ),
            "rt",
        ) as f:
            self.ast_mod = ast_parse(f.read(), skip_docstring_remit=True)

        self.func_node = next(
            filter(
                lambda func: func.name == "add1",
                filter(rpartial(isinstance, FunctionDef), walk(self.ast_mod)),
            )
        )

    def test_find_cst_at_ast(self) -> None:
        """Test that `find_cst_at_ast` can find the CST for one function"""
        cst_idx, cst_node = find_cst_at_ast(cstify_cst, self.func_node)
        self.assertIsNotNone(cst_idx)
        self.assertIsNotNone(cst_node)
        self.assertIsInstance(cst_node, FunctionDefinitionStart)

    def test_find_cst_at_ast_finds_all_functions(self) -> None:
        """Test that `find_cst_at_ast` can find the CST for all functions"""
        funcs = tuple(
            filter(rpartial(isinstance, FunctionDef), walk(self.ast_mod)),
        )
        for func in funcs:
            cst_idx, cst_node = find_cst_at_ast(cstify_cst, func)
            self.assertIsNotNone(cst_idx)
            self.assertIsNotNone(cst_node, "{name} not found".format(name=func.name))
            self.assertIsInstance(cst_node, FunctionDefinitionStart)

    def test_find_cst_at_ast_finds_class(self) -> None:
        """Test that `find_cst_at_ast` can find the CST for all functions"""
        class_def = next(
            filter(rpartial(isinstance, ClassDef), walk(self.ast_mod)),
        )
        cst_idx, cst_node = find_cst_at_ast(cstify_cst, class_def)
        self.assertIsNotNone(cst_idx)
        self.assertIsNotNone(cst_node)
        self.assertIsInstance(cst_node, ClassDefinitionStart)

    def test_find_cst_at_ast_errors_on_module(self) -> None:
        """Test that `find_cst_at_ast` fails to find the CST for `ast.Module`"""

        with patch("cdd.shared.ast_cst_utils.stderr", new_callable=StringIO) as e:
            find_cst_at_ast(cstify_cst, self.ast_mod)

        self.assertEqual(e.getvalue(), "`Module` not implemented\n")

    def maybe_replace_doc_str_in_function_or_class_test(
        self, existing_doc_str, new_doc_str, delta
    ) -> None:
        """
        Helper for writing `maybe_replace_doc_str_in_function_or_class` tests

        :param existing_doc_str: Existing doc_str (delete if None else replace)
        :type existing_doc_str: ```Optional[str]```

        :param new_doc_str: doc_str expected to end up with (delete if None else replace)
        :type new_doc_str: ```AST```

        :param delta: Delta value indicating what changed (if anything)
        :type delta: ```Delta```
        """

        cst_list = list(deepcopy(cstify_cst))
        cst_idx, cst_node = find_cst_at_ast(cst_list, self.func_node)
        self.assertIsNotNone(cst_node)
        if existing_doc_str is None:
            del cst_list[cst_idx + 1]
        else:
            cst_list[cst_idx + 1] = TripleQuoted(
                is_double_q=cst_list[cst_idx + 1].is_double_q,
                is_docstr=cst_list[cst_idx + 1].is_docstr,
                value=existing_doc_str,
                line_no_start=cst_list[cst_idx + 1].line_no_start,
                line_no_end=cst_list[cst_idx + 1].line_no_end,
            )

        func_node = deepcopy(self.func_node)
        func_node.body[0] = (
            Pass() if new_doc_str is None else Expr(set_value(new_doc_str))
        )

        with patch("cdd.shared.ast_cst_utils.debug_doctrans", identity):
            self.assertEqual(
                delta,
                maybe_replace_doc_str_in_function_or_class(
                    func_node, cst_idx, cst_list
                ),
            )

        if (
            isinstance(cst_list[cst_idx + 1], TripleQuoted)
            and cst_list[cst_idx + 1].is_docstr
        ):
            self.assertEqual(
                cst_list[cst_idx + 1].value.strip()[3:-3].strip(), new_doc_str
            )

    def test_maybe_replace_doc_str_in_function_or_class_replaced(self) -> None:
        """tests test_maybe_replace_doc_str_in_function_or_class succeeds in replacing"""

        self.maybe_replace_doc_str_in_function_or_class_test(
            existing_doc_str=(
                "\n"
                '        """\n'
                "        :param foo: a foo\n"
                "        :type foo: ```int```\n"
                "\n"
                "        :return: foo + 1\n"
                "        :rtype: ```int```\n"
                '        """'
            ),
            new_doc_str="Rewritten docstring",
            delta=Delta.replaced,
        )

    def test_maybe_replace_doc_str_in_function_or_class_added(self) -> None:
        """tests test_maybe_replace_doc_str_in_function_or_class succeeds in adding"""

        self.maybe_replace_doc_str_in_function_or_class_test(
            existing_doc_str=None,
            new_doc_str="New docstring",
            delta=Delta.added,
        )

    def test_maybe_replace_doc_str_in_function_or_class_removed(self) -> None:
        """tests test_maybe_replace_doc_str_in_function_or_class succeeds in removing"""

        self.maybe_replace_doc_str_in_function_or_class_test(
            existing_doc_str="Remove this",
            new_doc_str=None,
            delta=Delta.removed,
        )

    def maybe_replace_function_return_type_test(
        self,
        ast_node_to_find,
        ast_node_to_make,
        cur_ast_node,
        cur_cst_value,
        before,
        after,
        delta,
    ) -> None:
        """
        Helper for writing `maybe_replace_function_return_type` tests

        :param ast_node_to_find: AST node to find
        :type ast_node_to_find: ```AST```

        :param ast_node_to_make: AST node to end up with. If None
        :type ast_node_to_make: ```AST```

        :param cur_ast_node: AST parse of the CST node that currently exists where the `ast_node_to_find` is
        :type cur_ast_node: ```AST```

        :param cur_cst_value: Create a new FunctionDefinitionStart with this. If provided will replace existing.
        :type cur_cst_value: ```Optional[str]```

        :param before: Function prototype before function is run
        :type before: ```str```

        :param after: Function prototype after function is run
        :type after: ```str```

        :param delta: Delta value indicating what changed (if anything)
        :type delta: ```Delta```
        """

        cst_list = list(deepcopy(cstify_cst))
        cst_idx, cst_node = find_cst_at_ast(cst_list, ast_node_to_find)
        self.assertIsNotNone(cst_node)
        if cur_cst_value is not None:
            cst_list[cst_idx] = FunctionDefinitionStart(
                line_no_start=cst_list[cst_idx].line_no_start,
                line_no_end=cst_list[cst_idx].line_no_end,
                name=cst_list[cst_idx].name,
                value=cur_cst_value,
            )
        self.assertEqual(
            "".join(map(attrgetter("value"), cst_list[cst_idx : cst_idx + 1])), before
        )
        with patch("cdd.shared.ast_cst_utils.debug_doctrans", identity):
            self.assertEqual(
                maybe_replace_function_return_type(
                    new_node=ast_node_to_make,
                    cst_idx=cst_idx,
                    cst_list=cst_list,
                    cur_ast_node=cur_ast_node,
                ),
                delta,
            )
        self.assertEqual(
            "".join(map(attrgetter("value"), cst_list[cst_idx : cst_idx + 1])), after
        )

    def test_maybe_replace_function_return_type_adds(self) -> None:
        """
        Tests that `maybe_replace_function_return_type` adds return type
        """

        before = "\n\n    @staticmethod\n    def add1(foo):"
        after = "\n\n    @staticmethod\n    def add1(foo) -> int:"

        new_func_node = deepcopy(self.func_node)
        new_func_node.returns = Name("int", Load())

        self.maybe_replace_function_return_type_test(
            ast_node_to_find=self.func_node,
            ast_node_to_make=new_func_node,
            cur_ast_node=ast_parse(
                "{func_start} pass".format(func_start=before.strip().replace("  ", "")),
                skip_annotate=True,
                skip_docstring_remit=True,
            ).body[0],
            cur_cst_value=None,
            before=before,
            after=after,
            delta=Delta.added,
        )

    def test_maybe_replace_function_return_type_removes(self) -> None:
        """
        Tests that `maybe_replace_function_return_type` removes return type
        """

        before = "\n\n    @staticmethod\n    def add1(foo) -> int:"
        after = "\n\n    @staticmethod\n    def add1(foo):"

        self.maybe_replace_function_return_type_test(
            ast_node_to_find=self.func_node,
            ast_node_to_make=self.func_node,
            cur_ast_node=ast_parse(
                "{func_start} pass".format(func_start=before.strip().replace("  ", "")),
                skip_annotate=True,
                skip_docstring_remit=True,
            ).body[0],
            cur_cst_value=before,
            before=before,
            after=after,
            delta=Delta.removed,
        )

    def test_maybe_replace_function_return_type_replaces(self) -> None:
        """
        Tests that `maybe_replace_function_return_type` replaces return type
        """

        before = "\n\n    @staticmethod\n    def add1(foo) -> int:"
        after = "\n\n    @staticmethod\n    def add1(foo) -> float:"

        new_func_node = deepcopy(self.func_node)
        new_func_node.returns = Name("float", Load())

        self.maybe_replace_function_return_type_test(
            ast_node_to_find=self.func_node,
            ast_node_to_make=new_func_node,
            cur_ast_node=ast_parse(
                "{func_start} pass".format(func_start=before.strip().replace("  ", "")),
                skip_annotate=True,
                skip_docstring_remit=True,
            ).body[0],
            cur_cst_value=before,
            before=before,
            after=after,
            delta=Delta.replaced,
        )

    def test_maybe_replace_function_args_nop(self) -> None:
        """
        Tests that `maybe_replace_function_args` does nothing on equal args
        """

        func_src = "\n\n    @staticmethod\n    def add1(foo: int) -> int:"

        func_node = ast_parse(
            "{func_start} pass".format(func_start=func_src.strip().replace("  ", "")),
            skip_annotate=True,
            skip_docstring_remit=True,
        ).body[0]

        cst_list = list(deepcopy(cstify_cst))
        cst_idx, cst_node = find_cst_at_ast(cst_list, self.func_node)
        self.assertIsNotNone(cst_node)
        cst_list[cst_idx] = FunctionDefinitionStart(
            line_no_start=cst_list[cst_idx].line_no_start,
            line_no_end=cst_list[cst_idx].line_no_end,
            name=cst_list[cst_idx].name,
            value=func_src,
        )
        self.assertEqual(
            maybe_replace_function_args(
                new_node=func_node,
                cur_ast_node=func_node,
                cst_idx=cst_idx,
                cst_list=cst_list,
            ),
            Delta.nop,
        )
        self.assertEqual(
            "".join(map(attrgetter("value"), cst_list[cst_idx : cst_idx + 1])), func_src
        )

    def maybe_replace_function_args_test(self, before, after, delta):
        """
        Helper for writing `maybe_replace_function_args` tests

        :param before: Function prototype before function is run
        :type before: ```str```

        :param after: Function prototype after function is run
        :type after: ```str```

        :param delta: Delta value indicating what changed (if anything)
        :type delta: ```Delta```
        """

        new_node = ast_parse(
            reindent_block_with_pass_body(after),
            skip_annotate=True,
            skip_docstring_remit=True,
        ).body[0]

        cur_ast_node = ast_parse(
            reindent_block_with_pass_body(before),
            skip_annotate=True,
            skip_docstring_remit=True,
        ).body[0]

        cst_list = list(deepcopy(cstify_cst))
        cst_idx, cst_node = find_cst_at_ast(cst_list, self.func_node)
        self.assertIsNotNone(cst_node)
        cst_list[cst_idx] = FunctionDefinitionStart(
            line_no_start=cst_list[cst_idx].line_no_start,
            line_no_end=cst_list[cst_idx].line_no_end,
            name=cst_list[cst_idx].name,
            value=before,
        )
        with patch("cdd.shared.ast_cst_utils.debug_doctrans", identity):
            self.assertEqual(
                maybe_replace_function_args(
                    new_node=new_node,
                    cur_ast_node=cur_ast_node,
                    cst_idx=cst_idx,
                    cst_list=cst_list,
                ),
                delta,
            )
        self.assertEqual(
            after, "".join(map(attrgetter("value"), cst_list[cst_idx : cst_idx + 1]))
        )

    def test_maybe_replace_function_args_added(self) -> None:
        """
        Tests that `maybe_replace_function_args` adds to args (adds type annotations)
        """

        self.maybe_replace_function_args_test(
            before="\n\n    @staticmethod\n    def add1(foo) -> int:",
            after="\n\n    @staticmethod\n    def add1(foo: int) -> int:",
            delta=Delta.added,
        )

    def test_maybe_replace_function_args_removed(self) -> None:
        """
        Tests that `maybe_replace_function_args` removes from args (removes type annotations)
        """

        self.maybe_replace_function_args_test(
            before="\n\n    @staticmethod\n    def add1(foo: int) -> int:",
            after="\n\n    @staticmethod\n    def add1(foo) -> int:",
            delta=Delta.removed,
        )

    def test_maybe_replace_function_args_replaced(self) -> None:
        """
        Tests that `maybe_replace_function_args` replaces to args (replaces type annotations)
        """

        self.maybe_replace_function_args_test(
            before="\n\n    @staticmethod\n    def add1(foo: int) -> int:",
            after="\n\n    @staticmethod\n    def add1(foo: float) -> int:",
            delta=Delta.replaced,
        )

    def test_debug_doctrans(self) -> None:
        """
        Test that the `print` function is called the right number of times from  `debug_doctrans`' invocation
        """
        with patch(
            "cdd.shared.ast_cst_utils.print", new_callable=MagicMock()
        ) as print_func:
            debug_doctrans(Delta.added, "", "", "")
        self.assertEqual(print_func.call_count, 1)


unittest_main()



"""
Utility functions for `cdd.emit.class_`
"""

import ast
from ast import Attribute, Load, Name


class RewriteName(ast.NodeTransformer):
    """
    A :class:`NodeTransformer` subclass that walks the abstract syntax tree and
    allows modification of nodes. Here it modifies parameter names to be `self.param_name`
    """

    def __init__(self, node_ids):
        """
        Set parameter

        :param node_ids: Container of AST `id`s to match for rename
        :type node_ids: ```Optional[Iterator[str]]```
        """
        self.node_ids = node_ids

    def visit_Name(self, node):
        """
        Rename parameter name with a `self.` attribute prefix

        :param node: The AST node
        :type node: ```Name```

        :return: `Name` iff `Name` is not a parameter else `Attribute`
        :rtype: ```Union[Name, Attribute]```
        """
        return (
            Attribute(Name("self", Load()), node.id, Load())
            if not self.node_ids or node.id in self.node_ids
            else ast.NodeTransformer.generic_visit(self, node)
        )


__all__ = ["RewriteName"]



"""
Tests for the Intermediate Representation produced by the `pydantic` parser
"""

import ast
from unittest import TestCase

import cdd.pydantic.parse
from cdd.tests.mocks.ir import pydantic_ir
from cdd.tests.mocks.pydantic import pydantic_class_cls_def, pydantic_class_str
from cdd.tests.utils_for_tests import run_ast_test, unittest_main


class TestParsePydantic(TestCase):
    """
    Tests whether the intermediate representation is consistent when parsed from different inputs.

    IR is a dictionary of form:
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    """

    def test_from_pydantic_roundtrip(self) -> None:
        """
        Tests pydantic roundtrip of mocks
        """
        run_ast_test(
            self, ast.parse(pydantic_class_str).body[0], pydantic_class_cls_def
        )

    def test_from_pydantic(self) -> None:
        """
        Tests whether `pydantic` produces `pydantic_ir`
              from `pydantic_class_cls_def`
        """
        ir = cdd.pydantic.parse.pydantic(pydantic_class_cls_def)
        del ir["_internal"]  # Not needed for this test
        self.assertDictEqual(ir, pydantic_ir)


unittest_main()



"""
Utility functions for `cdd.emit.json_schema`
"""

import ast
from ast import AST, Set

from cdd.json_schema.utils.parse_utils import json_type2typ
from cdd.shared.ast_utils import Set_to_set, ast_type_to_python_type, get_value
from cdd.shared.pure_utils import none_types


def param2json_schema_property(param, required):
    """
    Turn a param into a JSON schema property

    :param param: Name, dict with keys: 'typ', 'doc', 'default'
    :type param: ```Tuple[str, dict]```

    :param required: Required parameters. This function may push to the list.
    :type required: ```List[str]```

    :return: JSON schema property. Also may push to `required`.
    :rtype: ```dict```
    """
    name, _param = param
    del param
    if _param.get("doc"):
        _param["description"] = _param.pop("doc")
    if _param.get("typ") == "datetime":
        del _param["typ"]
        _param.update({"type": "string", "format": "date-time"})
        required.append(name)
    elif _param.get("typ") in typ2json_type:
        _param["type"] = typ2json_type[_param.pop("typ")]
        required.append(name)
    elif _param.get("typ", ast) is not ast:
        _param["type"] = _param.pop("typ")
        if _param["type"].startswith("Optional["):
            _param["type"] = _param["type"][len("Optional[") : -1]
            if _param["type"] in typ2json_type:
                _param["type"] = typ2json_type[_param["type"]]
            # elif _param.get("typ") in typ2json_type:
            #    _param["type"] = typ2json_type[_param.pop("typ")]
        else:
            required.append(name)

        if _param["type"].startswith("Literal["):
            parsed_typ = get_value(ast.parse(_param["type"]).body[0])
            assert (
                parsed_typ.value.id == "Literal"
            ), "Only basic Literal support is implemented, not {}".format(
                parsed_typ.value.id
            )
            enum = sorted(map(get_value, get_value(parsed_typ.slice).elts))
            _param.update(
                {
                    "pattern": "|".join(enum),
                    "type": typ2json_type[type(enum[0]).__name__],
                }
            )
    if _param.get("default", False) in none_types:
        del _param["default"]  # Will be inferred as `null` from the type
    elif isinstance(_param.get("default"), AST):
        _param["default"] = ast_type_to_python_type(_param["default"])
    if isinstance(_param.get("choices"), Set):
        _param["pattern"] = "|".join(
            sorted(map(str, Set_to_set(_param.pop("choices"))))
        )
    return name, _param


typ2json_type = {v: k for k, v in json_type2typ.items()}

__all__ = ["param2json_schema_property"]
